# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch load entry for all algorithms."""

import json
import os

from neural_compressor.common.utils import FP8_QUANT, Mode, load_config_mapping, log_process
from neural_compressor.torch.quantization.config import (
    AutoRoundConfig,
    AWQConfig,
    FP8Config,
    GPTQConfig,
    HQQConfig,
    RTNConfig,
    TEQConfig,
)
from neural_compressor.torch.utils import SaveLoadFormat, get_enum_from_format, read_json_file

config_name_mapping = {
    FP8_QUANT: FP8Config,
}


@log_process(mode=Mode.SAVE)
def save(model, checkpoint_dir="saved_results", format="default"):
    """Save quantized model.

    Args:
        model (torch.nn.module or TorchScript model with IPEX or fx graph with pt2e, optional): Quantized model.
        checkpoint_dir (str, optional): checkpoint directory. Defaults to "saved_results".
        format (str, optional): 'default' for saving INC quantized model.
            'huggingface' for saving huggingface WOQ causal language model.
            'vllm' for saving FP8 model like 'neuralmagic/Meta-Llama-3.1-8B-Instruct-FP8'
            quantized by llm-compressor(https://github.com/vllm-project/llm-compressor).
            Defaults to "default".
    """
    format = get_enum_from_format(format)
    config_mapping = model.qconfig
    config_object = config_mapping[next(iter(config_mapping))]
    # fp8_quant
    if isinstance(config_object, FP8Config):
        from neural_compressor.torch.algorithms import fp8_quant

        if format == SaveLoadFormat.DEFAULT:
            format = SaveLoadFormat.HUGGINGFACE
        fp8_quant.save(model, checkpoint_dir, format)
    elif isinstance(config_object, (RTNConfig, GPTQConfig, AWQConfig, TEQConfig, AutoRoundConfig)):
        from neural_compressor.torch.algorithms import weight_only

        weight_only.save(model, checkpoint_dir, format=format)
    else:
        assert format == SaveLoadFormat.DEFAULT, "Currently, only default format is supported."
        model.save(checkpoint_dir)


@log_process(mode=Mode.LOAD)
def load(model_name_or_path, original_model=None, format="default", device="cpu", **kwargs):
    """Load quantized model.

    1. Load INC quantized model in local.
        case 1: WOQ
            from neural_compressor.torch.quantization import load
            load(model_name_or_path="saved_results", original_model=fp32_model)

        case 2: INT8/FP8
            from neural_compressor.torch.quantization import load
            load(model_name_or_path='saved_result', original_model=fp32_model)

        case 3: TorchScript (IPEX)
            from neural_compressor.torch.quantization import load
            load(model_name_or_path='saved_result')

    2. Load HuggingFace quantized model, including GPTQ models and upstreamed INC quantized models in HF model hub.
        case 1: WOQ
            from neural_compressor.torch.quantization import load
            load(model_name_or_path=model_name_or_path, format="huggingface")

    Args:
        model_name_or_path (str):  torch checkpoint directory or hugginface model_name_or_path.
            If 'format' is set to 'huggingface', it means the huggingface model_name_or_path.
            If 'format' is set to 'default', it means the 'checkpoint_dir'.
            Parameter should not be None. it co-works with 'original_model' parameter to load INC
            quantized model in local.
        original_model (torch.nn.module or TorchScript model with IPEX or fx graph with pt2e, optional):
            original model before quantization. Needed if 'format' is set to 'default' and not TorchScript model.
            Defaults to None.
        format (str, optional): 'default' for loading INC quantized model.
            'huggingface' for loading huggingface WOQ causal language model. Defaults to "default".
        device (str, optional): 'cpu', 'hpu'. specify the device the model will be loaded to.
            currently only used for weight-only quantization.
        kwargs (remaining dictionary of keyword arguments, optional):
            remaining dictionary of keyword arguments for loading huggingface models.
            Will be passed to the huggingface model's `__init__` method, such as 'trust_remote_code', 'revision'.

    Returns:
        The quantized model
    """
    format = get_enum_from_format(format)
    if format == SaveLoadFormat.DEFAULT:
        from neural_compressor.common.base_config import ConfigRegistry

        qconfig_file_path = os.path.join(os.path.abspath(os.path.expanduser(model_name_or_path)), "qconfig.json")
        if not os.path.exists(qconfig_file_path):
            raise ValueError("qconfig.json file is necessary for the default format.")
        with open(qconfig_file_path, "r") as f:
            per_op_qconfig = json.load(f)

        if " " in per_op_qconfig.keys():  # ipex qconfig format: {' ': {'q_op_infos': {'0': {'op_type': ...
            from neural_compressor.torch.algorithms import static_quant

            return static_quant.load(model_name_or_path)
        elif "static_quant" in per_op_qconfig.keys() or "pt2e_dynamic_quant" in per_op_qconfig.keys():  # PT2E
            from neural_compressor.torch.algorithms import pt2e_quant

            return pt2e_quant.load(model_name_or_path)
        else:
            config_mapping = load_config_mapping(qconfig_file_path, ConfigRegistry.get_all_configs()["torch"])
            # select load function
            config_object = config_mapping[next(iter(config_mapping))]

            if isinstance(
                config_object, (RTNConfig, GPTQConfig, AWQConfig, TEQConfig, AutoRoundConfig, HQQConfig)
            ):  # WOQ
                from neural_compressor.torch.algorithms import weight_only

                qmodel = weight_only.load(
                    model_name_or_path, original_model, format=SaveLoadFormat.DEFAULT, device=device, **kwargs
                )
                return qmodel.to(device)
    elif format == SaveLoadFormat.HUGGINGFACE:
        import transformers

        try:
            config = transformers.AutoConfig.from_pretrained(model_name_or_path, **kwargs)
            quantization_config = config.quantization_config
        except:
            quantization_config_file = "quantization_config.json"
            # for Flux pipeline
            if os.path.exists(model_name_or_path):
                # If the model_name_or_path is a local path, try to load the config from there
                quantization_config_path = os.path.join(model_name_or_path, quantization_config_file)
            else:
                # If the model_name_or_path is a Hugging Face model ID, try to download the config
                from huggingface_hub import hf_hub_download

                quantization_config_path = hf_hub_download(
                    repo_id=model_name_or_path,
                    filename=quantization_config_file,
                    revision=kwargs.get("revision", "main"),
                )
            quantization_config = read_json_file(quantization_config_path)
            kwargs["quantization_config"] = quantization_config

        if original_model is not None:
            kwargs["original_model"] = original_model
        # use config to check which algorithm is used.
        if (
            "fp8_config" in quantization_config
            or
            # for FP8 LLMs for vLLM (https://huggingface.co/neuralmagic).
            (
                "quant_method" in quantization_config
                and quantization_config["quant_method"] in ["fp8", "compressed-tensors"]
            )
        ):
            from neural_compressor.torch.algorithms import fp8_quant

            return fp8_quant.load(model_name_or_path, format=format, device=device, **kwargs)
        else:
            from neural_compressor.torch.algorithms import weight_only

            qmodel = weight_only.load(model_name_or_path, format=SaveLoadFormat.HUGGINGFACE, device=device, **kwargs)
            return qmodel.to(device)
    else:
        assert False, "Unexpected format: {} occurred during model loading".format(format)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Rewrite the FP32 operators to FP16 or BF16 operators."""

from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Dict, List, Tuple

import torch
import torch.ao.quantization.quantizer.x86_inductor_quantizer as xiq
import torch.ao.quantization.quantizer.xnnpack_quantizer as xpq
from torch.fx import subgraph_rewriter
from torch.fx.experimental.proxy_tensor import make_fx
from torch.fx.subgraph_rewriter import Match
from typing_extensions import TypeAlias

from neural_compressor.common import logger, utils

# =============================================================================
# Search and replace patterns
# =============================================================================
TorchFuncType: TypeAlias = Callable[..., Any]


@dataclass
class PatternPair:
    """Represents a pair of patterns used for search and replacement in a graph.

    Attributes:
        fn (TorchFuncType): The function type associated with the pattern pair.
        search_pattern (torch.fx.GraphModule): The search pattern to be matched in the graph.
        replace_pattern (torch.fx.GraphModule): The replacement pattern to be used when a match is found.
    """

    fn: TorchFuncType
    search_pattern: torch.fx.GraphModule
    replace_pattern: torch.fx.GraphModule


# key: torch func
# value: the tuple of args
FuncArgsMappingType: TypeAlias = Dict[TorchFuncType, List[Tuple[torch.Tensor, ...]]]


# Align with xiq, as it relay on xiq's set_module_xx capability
FN_ARGS_MAPPING: FuncArgsMappingType = {
    # Note: ORDER is matter
    torch.nn.functional.linear: [
        (torch.randn(0, 0), torch.randn(0, 0)),  # linear w/o bias
        (torch.randn(0, 0), torch.randn(0, 0), torch.randn(0)),  # linear w/ bias
    ],
    torch.nn.functional.conv2d: [
        (torch.randn(1, 1, 1, 1), torch.randn(1, 1, 1, 1)),  # conv2d w/o bias
        (torch.randn(1, 1, 1, 1), torch.randn(1, 1, 1, 1), torch.randn(1)),  # conv2d w/ bias
    ],
    torch.matmul: [
        (torch.randn(0, 0), torch.randn(0, 0)),
        (torch.randn(0, 0, 0), torch.randn(0, 0, 0)),
        (torch.randn(0, 0, 0, 0), torch.randn(0, 0, 0, 0)),
    ],
}

# module cls <-> function name
NN_MODULES_TO_NN_FN = {
    torch.nn.Linear: torch.nn.functional.linear,
    torch.nn.Conv2d: torch.nn.functional.conv2d,
}

# Use the mapping from xiq
FN_ATEN_OPS_MAPPING = xiq._map_module_function_to_aten_operator_type()

SUPPORTED_OPERATORS = FN_ATEN_OPS_MAPPING.values()


PatternRegistryType: TypeAlias = Dict[TorchFuncType, PatternPair]
HALF_PRECISION_PATTERN_REGISTRY: Dict[torch.dtype, PatternRegistryType] = {
    torch.float16: defaultdict(list),
    torch.bfloat16: defaultdict(list),
}

# FP16_PATTERN_REGISTRY: PatternRegistryType = HALF_PRECISION_PATTERN_REGISTRY[torch.float16]
# BF16_PATTERN_REGISTRY: PatternRegistryType = HALF_PRECISION_PATTERN_REGISTRY[torch.bfloat16]


def pattern_factory(fn: TorchFuncType, fn_arg: Tuple[torch.Tensor, ...], target_dtype: torch.dtype = torch.float16):
    """Create a search, replace pattern and filter functions for a given torch function and its arguments."""
    assert target_dtype in [
        torch.float16,
        torch.bfloat16,
    ], f"target_dtype should either be `torch.float16` or `torch.bfloat16`, but got {target_dtype}"

    def replace_fn_wrapper(fn_args, fn):
        converted_args = [arg.to(target_dtype) for arg in fn_args]
        target_dtype_out = fn(*converted_args)
        return target_dtype_out.float()

    replace_fn = partial(replace_fn_wrapper, fn=fn)

    search_pattern_gm = make_fx(fn, pre_dispatch=True)(*fn_arg)
    # TODO: double-check `*fn_args` or `fn_args`
    replace_pattern_gm = make_fx(replace_fn, pre_dispatch=True)(fn_arg)

    pattern_pair = PatternPair(fn, search_pattern_gm, replace_pattern_gm)

    return pattern_pair


def _register_pattern_pair(dtype: torch.dtype) -> None:
    for fn, fn_args_lst in FN_ARGS_MAPPING.items():
        for fn_args in fn_args_lst:
            logger.debug(f"Registering search and replace patterns for {fn} with args: {fn_args}.")
            pattern_pair = pattern_factory(fn, fn_args)
            HALF_PRECISION_PATTERN_REGISTRY[dtype][fn].append(pattern_pair)
    utils.logger.debug(
        f"Registered {len(HALF_PRECISION_PATTERN_REGISTRY[dtype])} search and replace patterns for {dtype}."
    )


_register_pattern_pair(torch.float16)
_register_pattern_pair(torch.bfloat16)


def get_filter_fn(node_list, fn):
    """Filter function to check if a node with the target operator is in the given `node_list`.

    Args:
        node_list (list): List of nodes to check against.
        fn (str): Target operator.

    Returns:
        bool: True if the node with the target operator is in the `node_list`, False otherwise.
    """
    target_op = FN_ATEN_OPS_MAPPING[fn]

    def is_target_node_in_candidate_list(match, original_graph, pattern_graph):
        """Filter the node with target operator in match and check if it is in `node_list`."""
        target_node = None
        for node in pattern_graph.nodes:  # pragma: no cover
            if node.target == target_op:
                target_node = node
                break
        if target_node is None:  # pragma: no cover
            return False
        matched_node = match.nodes_map[target_node]
        return matched_node in node_list

    return is_target_node_in_candidate_list


def apply_single_pattern_pair(gm: torch.fx.GraphModule, pattern_pair: PatternPair, node_list):
    """Applies a single pattern pair to a given GraphModule.

    Args:
        gm (torch.fx.GraphModule): The GraphModule to apply the pattern pair to.
        pattern_pair (PatternPair): The pattern pair containing the search and replace patterns.
        node_list: The list of nodes to filter for pattern matching.

    Returns:
        List[Match]: A list of Match objects representing the matches found after applying the pattern pair.
    """
    filter_fn = get_filter_fn(node_list, pattern_pair.fn)
    match_and_replacements = subgraph_rewriter.replace_pattern_with_filters(
        gm=gm,
        pattern=pattern_pair.search_pattern,
        replacement=pattern_pair.replace_pattern,
        match_filters=[filter_fn],
    )
    utils.logger.info(f"Found {len(match_and_replacements)} matches.")

    match_list = [Match(anchor=m.anchor, nodes_map=m.nodes_map) for m in match_and_replacements]
    return match_list


def get_unquantized_node_set(gm: torch.fx.GraphModule):
    """Retrieves the set of unquantized nodes from a given GraphModule.

    Args:
        gm (torch.fx.GraphModule): The GraphModule to retrieve unquantized nodes from.

    Returns:
        set: A set containing the unquantized nodes.
    """
    unquantized_node_set = set()
    for node in gm.graph.nodes:
        if meta := getattr(node, "meta"):
            if quantization_annotation := meta.get(xiq.QUANT_ANNOTATION_KEY):
                none_annotation = xiq._X86InductorQuantizationAnnotation(_annotated=True)
                if quantization_annotation != none_annotation:  # pragma: no cover
                    continue
        unquantized_node_set.add(node)
    return unquantized_node_set


def transformation(gm: torch.fx.GraphModule, node_candidate_list: List[str], target_dtype: torch.dtype = torch.float16):
    """Convert the nodes in `node_candidate_list` to `target_dtype` if possible."""
    for pattern_pair_lst in HALF_PRECISION_PATTERN_REGISTRY[target_dtype].values():
        for pattern_pair in pattern_pair_lst:
            apply_single_pattern_pair(gm, pattern_pair, node_candidate_list)
    utils.logger.info(f"Half precision conversion({target_dtype}) completed.")
    if utils.level_name == "DEBUG":  # pragma: no cover
        gm.print_readable(True)


# =============================================================================
# Utils to parse the node candidate set for half precision conversion
# =============================================================================


def _parse_node_candidate_set_from_user_config(config, gm):
    """Parse the node candidate set from user config."""
    op_type_configs, op_name_configs = config._get_op_name_op_type_config()
    op_type_filters = []
    op_name_filters = []
    for op_type_name, config in op_type_configs.items():  # pragma: no cover
        op_type = getattr(torch.nn, op_type_name)
        if config.act_dtype in ["fp16", "bf16"]:  # pragma: no cover
            filter = xpq._get_module_type_filter(op_type)
            op_type_filters.append(filter)
    for op_name, config in op_name_configs.items():
        if config.act_dtype in ["fp16", "bf16"]:  # pragma: no cover
            filter = xpq._get_module_name_filter(op_name)
            op_name_filters.append(filter)
    node_set_from_user_config = set()
    all_filters = op_type_filters + op_name_filters
    for node in gm.graph.nodes:  # pragma: no cover
        if any([filter(node) for filter in all_filters]):
            node_set_from_user_config.add(node)
    return node_set_from_user_config


def get_half_precision_node_set(gm, config):
    """Retrieves a set of nodes from the given graph model (gm) that are candidates for conversion to half precision.

    The result is the intersection between `unquantized_node_set` and `node_set_from_user_config`.

    Args:
        gm (GraphModel): The graph model to search for nodes.
        config (dict): User configuration for node candidate set.

    Returns:
        set: A set of nodes that are candidates for conversion to half precision.
    """
    # TODO: implement it, current return all unquantized_node_set

    node_set_from_user_config = _parse_node_candidate_set_from_user_config(config, gm)
    unquantized_node_set = get_unquantized_node_set(gm)
    possible_node_set = unquantized_node_set.intersection(node_set_from_user_config)
    half_precision_node_set = set()
    for node in possible_node_set:
        if node.target in SUPPORTED_OPERATORS:
            half_precision_node_set.add(node)
    utils.logger.info(
        f"Found {len(half_precision_node_set)} nodes to convert to half precision: {half_precision_node_set}"
    )
    return half_precision_node_set



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for PT2E quantization."""

from typing import Dict

import torch
import torch.ao.quantization.quantizer.x86_inductor_quantizer as xiq
from torch.ao.quantization.observer import (
    HistogramObserver,
    MinMaxObserver,
    PerChannelMinMaxObserver,
    PlaceholderObserver,
)
from torch.ao.quantization.quantizer import QuantizationSpec
from torch.ao.quantization.quantizer.x86_inductor_quantizer import QuantizationConfig, X86InductorQuantizer

from neural_compressor.torch.utils import GT_OR_EQUAL_TORCH_VERSION_2_5, logger


def create_quant_spec_from_config(dtype, sym, granularity, algo, is_dynamic=False) -> QuantizationSpec:
    """Create a quantization specification based on the given configuration.

    Args:
        dtype (str): The desired data type for quantization. Valid options are "int8" and "uint8".
        sym (bool): Whether to use symmetric quantization or not.
        granularity (str): The granularity of quantization. Valid options are "per_channel" and "per_tensor".
        algo (str): The algorithm to use for quantization. Valid options are "placeholder", "minmax", and "kl".
        is_dynamic (bool, optional): Whether to use dynamic quantization or not. Defaults to False.

    Returns:
        QuantizationSpec: The created quantization specification.
    """
    dtype_mapping: Dict[str, torch.dtype] = {"int8": torch.int8, "uint8": torch.uint8}
    select_dtype = dtype_mapping[dtype]
    min_max_mapping = {torch.int8: (-128, 127), torch.uint8: (0, 255)}
    qscheme_mapping = {
        "per_channel": {True: torch.per_channel_symmetric, False: torch.per_tensor_affine},
        "per_tensor": {True: torch.per_tensor_symmetric, False: torch.per_tensor_affine},
    }
    observer_mapping = {
        "placeholder": PlaceholderObserver,
        "minmax": MinMaxObserver,
        "kl": HistogramObserver,
        "per_channel_minmax": PerChannelMinMaxObserver,
    }
    # Force to use placeholder observer for dynamic quantization
    if is_dynamic:
        algo = "placeholder"
    if f"{granularity}_{algo}" in observer_mapping:
        observer_or_fake_quant_ctr = observer_mapping[f"{granularity}_{algo}"]
    else:
        observer_or_fake_quant_ctr = observer_mapping[algo]
    # qscheme
    qscheme = qscheme_mapping[granularity][sym]
    quantization_spec = QuantizationSpec(
        dtype=select_dtype,
        quant_min=min_max_mapping[select_dtype][0],
        quant_max=min_max_mapping[select_dtype][1],
        observer_or_fake_quant_ctr=observer_or_fake_quant_ctr,
        ch_axis=0,
        qscheme=qscheme,
        is_dynamic=is_dynamic,
    )
    return quantization_spec


def _map_inc_config_to_torch_quant_config(inc_config, is_dynamic=False) -> QuantizationConfig:
    NOT_QUANT_DTYPES = ["fp32", "fp16", "bf16"]
    if inc_config.act_dtype in NOT_QUANT_DTYPES and inc_config.w_dtype in NOT_QUANT_DTYPES:  # pragma: no cover
        logger.debug("Got non-quantizable data types, skipping quantization.")
        return None
    default_quant_config = xiq.get_default_x86_inductor_quantization_config(is_dynamic=is_dynamic)
    input_act_quant_spec = create_quant_spec_from_config(
        inc_config.act_dtype, inc_config.act_sym, inc_config.act_granularity, inc_config.act_algo, is_dynamic=is_dynamic
    )
    weight_quant_spec = create_quant_spec_from_config(
        inc_config.w_dtype, inc_config.w_sym, inc_config.w_granularity, inc_config.w_algo
    )
    quant_config = QuantizationConfig(
        input_activation=input_act_quant_spec,
        output_activation=default_quant_config.output_activation,
        weight=weight_quant_spec,
        bias=default_quant_config.bias,
        is_qat=False,
    )
    return quant_config


def create_xiq_quantizer_from_pt2e_config(config, is_dynamic=False) -> X86InductorQuantizer:
    """Creates an instance of X86InductorQuantizer based on the given configuration.

    Args:
        config: The configuration object containing the quantization settings.
        is_dynamic: A boolean indicating whether dynamic quantization is enabled.

    Returns:
        An instance of X86InductorQuantizer initialized with the provided configuration.
    """
    quantizer = xiq.X86InductorQuantizer()
    # set global
    global_config = _map_inc_config_to_torch_quant_config(config, is_dynamic)
    quantizer.set_global(global_config)
    # need torch >= 2.5
    if GT_OR_EQUAL_TORCH_VERSION_2_5:  # pragma: no cover
        op_type_config_dict, op_name_config_dict = config._get_op_name_op_type_config()
        if op_type_config_dict:
            for op_type, config in op_type_config_dict.items():
                _nn_module_type = getattr(torch.nn, op_type, None)
                if _nn_module_type:
                    quantizer.set_module_type_qconfig(
                        _nn_module_type, _map_inc_config_to_torch_quant_config(config, is_dynamic)
                    )
                _nn_func_type = getattr(torch.nn.functional, op_type, None)
                if _nn_func_type:
                    quantizer.set_function_type_qconfig(
                        _nn_module_type, _map_inc_config_to_torch_quant_config(config, is_dynamic)
                    )
        if op_name_config_dict:
            for op_name, config in op_name_config_dict.items():
                quantizer.set_module_name_qconfig(op_name, _map_inc_config_to_torch_quant_config(config, is_dynamic))
    return quantizer



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch quantization API."""

from neural_compressor.torch.quantization.quantize import quantize, prepare, convert, finalize_calibration
from neural_compressor.torch.quantization.config import (
    RTNConfig,
    get_default_rtn_config,
    get_default_double_quant_config,
    GPTQConfig,
    get_default_gptq_config,
    AWQConfig,
    get_default_awq_config,
    INT8StaticQuantConfig,
    StaticQuantConfig,
    get_default_static_config,
    SmoothQuantConfig,
    get_default_sq_config,
    TEQConfig,
    get_default_teq_config,
    AutoRoundConfig,
    get_default_AutoRound_config,
    HQQConfig,
    get_default_hqq_config,
    FP8Config,
    get_default_fp8_config,
    get_default_fp8_config_set,
    MXQuantConfig,
    get_default_mx_config,
    MixedPrecisionConfig,
    get_default_mixed_precision_config,
    get_default_mixed_precision_config_set,
    get_woq_tuning_config,
    DynamicQuantConfig,
    get_default_dynamic_config,
    HybridGPTQConfig,
)

from neural_compressor.torch.quantization.autotune import (
    autotune,
    TuningConfig,
    get_all_config_set,
    get_rtn_double_quant_config_set,
)

### Quantization Function Registration ###
import neural_compressor.torch.quantization.algorithm_entry
from neural_compressor.torch.quantization.save_load_entry import save, load



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The quantizer using SmoothQuant path."""


import json
import os

import torch

try:
    import intel_extension_for_pytorch as ipex
except:  # pragma: no cover
    assert False, "Please install IPEX for smooth quantization."

from collections import OrderedDict
from types import MethodType

from packaging.version import Version

from neural_compressor.torch.algorithms import Quantizer

from .utility import (
    CpuInfo,
    TorchSmoothQuant,
    cfg_to_qconfig,
    dump_model_op_stats,
    get_ipex_version,
    get_quantizable_ops_recursively,
    ipex_config_path,
    logger,
    simple_inference,
    update_sq_scale,
)

ipex_ver = get_ipex_version()


class SmoothQuantQuantizer(Quantizer):
    """SmoothQuantQuantizer Class."""

    def __init__(self, quant_config: OrderedDict = {}):  # pragma: no cover
        """Init a SmoothQuantQuantizer object.

        Args:
            quant_config (OrderedDict, optional): quantization config for ops. Defaults to {}.
        """
        super().__init__(quant_config)

    def prepare(self, model, example_inputs, inplace=True, *args, **kwargs):
        """Prepares a given model for quantization.

        Args:
            model (torch.nn.Module): raw fp32 model or prepared model.
            example_inputs (tensor/tuple/dict): used to trace torch model.
            inplace (bool, optional): whether to carry out model transformations in-place. Defaults to True.

        Returns:
            A prepared model.
        """
        assert example_inputs is not None, "Please provide example_inputs for smooth quantization."
        assert not ipex_ver.release < Version("2.1").release, "IPEX version >= 2.1 is required for SmoothQuant."

        # Note: we should make sure smoothquant is only executed once with inplacing fp32 model.
        if hasattr(model, "_smoothquant_optimized") and model._smoothquant_optimized:  # pragma: no cover
            logger.info("The model is already optimized by SmoothQuant algorithm, skip it.")
            return model

        cfgs, op_infos_from_cfgs, output_tensor_id_op_name = (
            model.cfgs,
            model.op_infos_from_cfgs,
            model.output_tensor_id_op_name,
        )

        # check smoothquant alpha and act_algo value
        recipe_cfgs = self.quant_config.get("recipe_cfgs", None)
        alpha = recipe_cfgs["smooth_quant_args"]["alpha"]

        # Update json file in ipex_config_path
        cfg_to_qconfig(self.quant_config, cfgs, op_infos_from_cfgs, output_tensor_id_op_name, alpha, smooth_quant=True)
        model.eval()

        for op, _ in self.quant_config["op"].items():
            act_algo = self.quant_config["op"][op]["activation"]["algorithm"]

        # Check save_qconf_summary part is a workaround for IPEX bug.
        # Sometimes the prepared model from get_op_capablitiy loss this attribute.
        if not hasattr(model, "save_qconf_summary") or not hasattr(model, "load_qconf_summary"):  # pragma: no cover
            from torch.ao.quantization.observer import MinMaxObserver

            if ipex_ver.release >= Version("2.1.1").release:
                static_qconfig = ipex.quantization.get_smooth_quant_qconfig_mapping(
                    alpha=alpha, act_observer=MinMaxObserver
                )
            else:  # pragma: no cover
                if act_algo == "minmax":
                    static_qconfig = ipex.quantization.get_smooth_quant_qconfig_mapping(
                        alpha=alpha, act_observer=MinMaxObserver()
                    )
                    logger.warning(
                        "The int8 model accuracy will be close to 0 with MinMaxobserver, "
                        + "the suggested IPEX version is higher or equal than 2.1.100+cpu."
                    )
                else:
                    static_qconfig = ipex.quantization.get_smooth_quant_qconfig_mapping(alpha=alpha)

            if isinstance(example_inputs, dict):
                model = ipex.quantization.prepare(
                    model, static_qconfig, example_kwarg_inputs=example_inputs, inplace=inplace
                )
            else:
                model = ipex.quantization.prepare(model, static_qconfig, example_inputs=example_inputs, inplace=inplace)

        model.load_qconf_summary(qconf_summary=ipex_config_path)
        return model

    def convert(self, model, example_inputs, inplace=True, *args, **kwargs):
        """Converts a prepared model to a quantized model.

        Args:
            model (QuantizationInterceptionModule): the prepared model to be converted.
            example_inputs (tensor/tuple/dict): used to trace torch model.
            inplace (bool, optional): whether to carry out model transformations in-place. Defaults to True.

        Returns:
            A quantized model.
        """
        use_bf16 = self.quant_config.get("use_bf16", None)

        model.save_qconf_summary(qconf_summary=ipex_config_path)
        model = _ipex_post_quant_process(model, example_inputs, use_bf16, inplace=inplace)

        with open(ipex_config_path, "r") as f:
            model.tune_cfg = json.load(f)
        model.ipex_config_path = ipex_config_path
        dump_model_op_stats(self.quant_config["op"])

        from neural_compressor.torch.algorithms.smooth_quant import save

        logger.info("Smooth quantization done.")
        model.ori_save = model.save
        model.save = MethodType(save, model)
        return model

    def quantize(self, model, tune_cfg, run_fn, example_inputs, inplace=True, *args, **kwargs):
        """Executes the quantize process on the specified model.

        Args:
            model (torch.nn.Module): raw fp32 model or prepared model.
            tune_cfg (OrderedDict): quantization config for ops.
            run_fn (Callable): a calibration function for calibrating the model.
            example_inputs (tensor/tuple/dict): used to trace torch model.
            inplace (bool, optional): whether to carry out model transformations in-place. Defaults to True.

        Returns:
            A quantized model.
        """
        assert not ipex_ver.release < Version("2.1").release, "IPEX version >= 2.1 is required for SmoothQuant."

        cfgs, op_infos_from_cfgs, output_tensor_id_op_name = (
            model.cfgs,
            model.op_infos_from_cfgs,
            model.output_tensor_id_op_name,
        )

        use_bf16 = tune_cfg.get("use_bf16", None)

        # check smoothquant folding value
        recipe_cfgs = tune_cfg.get("recipe_cfgs", None)
        if "smooth_quant_args" in recipe_cfgs and "folding" in recipe_cfgs["smooth_quant_args"]:
            if recipe_cfgs["smooth_quant_args"]["folding"] is None:  # pragma: no cover
                if ipex_ver.release < Version("2.1").release:
                    folding = True
                else:
                    folding = False
            else:
                folding = recipe_cfgs["smooth_quant_args"]["folding"]

        # Note: we should make sure smoothquant is only executed once with inplacing fp32 model.
        if hasattr(model, "_smoothquant_optimized") and model._smoothquant_optimized:  # pragma: no cover
            logger.info("The model is already optimized by SmoothQuant algorithm, skip it.")
            return model

        sq_info = model.sq_info

        # Update model parameter when smoothquant folding = False
        if recipe_cfgs and recipe_cfgs.get("smooth_quant", False) and not folding:
            return qdq_quantize(
                model,
                tune_cfg,
                run_fn,
                example_inputs,
                inplace,
                cfgs,
                op_infos_from_cfgs,
                output_tensor_id_op_name,
                sq_info,
            )

        # Update model parameter when smoothquant folding = True
        if recipe_cfgs and recipe_cfgs.get("smooth_quant", False) and folding:
            _apply_pre_optimization(model, tune_cfg, sq_info)

        # Update json file in ipex_config_path
        cfg_to_qconfig(self.quant_config, cfgs, op_infos_from_cfgs, output_tensor_id_op_name)
        model.eval()

        # Check save_qconf_summary part is a workaround for IPEX bug.
        # Sometimes the prepared model from get_op_capablitiy loss this attribute
        if not hasattr(model, "save_qconf_summary") or not hasattr(model, "load_qconf_summary"):  # pragma: no cover
            static_qconfig = ipex.quantization.default_static_qconfig_mapping
            if isinstance(example_inputs, dict):
                model = ipex.quantization.prepare(
                    model, static_qconfig, example_kwarg_inputs=example_inputs, inplace=inplace
                )
            else:
                model = ipex.quantization.prepare(model, static_qconfig, example_inputs=example_inputs, inplace=inplace)

        run_fn(model)
        model.save_qconf_summary(qconf_summary=ipex_config_path)
        model = _ipex_post_quant_process(model, example_inputs, use_bf16, inplace=inplace)

        # Recover model parameter when smoothquant folding = True
        if (
            recipe_cfgs
            and recipe_cfgs.get("smooth_quant", False)
            and recipe_cfgs["smooth_quant_args"]["folding"]
            and not inplace
        ):  # pragma: no cover
            _apply_pre_optimization(model, tune_cfg, sq_info, recover=True)

        with open(ipex_config_path, "r") as f:
            model.tune_cfg = json.load(f)
        model.ipex_config_path = ipex_config_path
        dump_model_op_stats(tune_cfg["op"])

        from neural_compressor.torch.algorithms.smooth_quant import save

        logger.info("Smooth quantization done.")
        model.ori_save = model.save
        model.save = MethodType(save, model)
        return model


def qdq_quantize(
    model, tune_cfg, run_fn, example_inputs, inplace, cfgs, op_infos_from_cfgs, output_tensor_id_op_name, sq
):
    """Executes the smooth quantize process.

    Args:
        model (torch.nn.Module): raw fp32 model or prepared model.
        tune_cfg (OrderedDict): quantization config for ops.
        run_fn (Callable): a calibration function for calibrating the model.
        example_inputs (tensor/tuple/dict): used to trace torch model.
        inplace (bool): whether to carry out model transformations in-place. Defaults to True.
        cfgs (dict): configs loaded from ipex config path.
        op_infos_from_cfgs (dict): dict containing configs that have been parsed for each op.
        output_tensor_id_op_name (dict): dict containing op names corresponding to 'op_infos_from_cfgs'.
        sq (TorchSmoothQuant): TorchSmoothQuant class containing sq infos.

    Returns:
        A quantized model.
    """
    smoothquant_scale_info = sq.sq_scale_info
    sq_minmax_init = True if tune_cfg.get("act_algo", "kl") == "minmax" else False

    use_bf16 = tune_cfg.get("use_bf16", None)

    # Check save_qconf_summary part is a workaround for IPEX bug.
    # Sometimes the prepared model from get_op_capablitiy loss this attribute
    if not hasattr(model, "save_qconf_summary") or not hasattr(model, "load_qconf_summary"):  # pragma: no cover
        from torch.ao.quantization.observer import MinMaxObserver

        if ipex_ver.release >= Version("2.1.1").release:
            static_qconfig = ipex.quantization.get_smooth_quant_qconfig_mapping(alpha=0.5, act_observer=MinMaxObserver)
        else:  # pragma: no cover
            if sq_minmax_init:
                static_qconfig = ipex.quantization.get_smooth_quant_qconfig_mapping(
                    alpha=0.5, act_observer=MinMaxObserver()
                )
                logger.warning(
                    "The int8 model accuracy will be close to 0 with MinMaxobserver, "
                    + "the suggested IPEX version is higher or equal than 2.1.100+cpu."
                )
            else:
                static_qconfig = ipex.quantization.get_smooth_quant_qconfig_mapping(alpha=0.5)
        if isinstance(example_inputs, dict):
            model = ipex.quantization.prepare(
                model, static_qconfig, example_kwarg_inputs=example_inputs, inplace=inplace
            )
        else:
            model = ipex.quantization.prepare(model, static_qconfig, example_inputs=example_inputs, inplace=inplace)

    # The IPEX SmoothQuant observer can only use save/load_qconf_summary once.
    # The save_qconf_summary API will freeze the scale used in model and calibration won't work anymore.
    # The load_qconf_summary will overwrite the scales used in model but only work in the first call.
    # Here, we use INC collected scale for Linear and set normal observer instead of SQObserver \
    # to make sure calibration works for other ops, like add, bmm.
    cfg_to_qconfig(tune_cfg, cfgs, op_infos_from_cfgs, output_tensor_id_op_name, smooth_quant=True)
    update_sq_scale(ipex_config_path, smoothquant_scale_info)
    model.load_qconf_summary(qconf_summary=ipex_config_path)
    # real calibration for other operators
    try:
        # IPEX may raise an error on the second iteration.
        # OverflowError: cannot convert float infinity to integer
        run_fn(model)
    except:  # pragma: no cover
        logger.warning(
            "The calibration failed when calibrating with ipex, "
            + "using scale info from SmoothQuant for Linear and "
            + "one iter calibration for other ops."
        )

    model.save_qconf_summary(qconf_summary=ipex_config_path)
    if ipex_ver.release > Version("2.1.0").release:
        update_sq_scale(ipex_config_path, smoothquant_scale_info)
        model.load_qconf_summary(qconf_summary=ipex_config_path)
    model = _ipex_post_quant_process(model, example_inputs, use_bf16, inplace=inplace)

    with open(ipex_config_path, "r") as f:
        model.tune_cfg = json.load(f)
    model.ipex_config_path = ipex_config_path
    dump_model_op_stats(tune_cfg["op"])

    from neural_compressor.torch.algorithms.smooth_quant import save

    logger.info("Smooth quantization done.")
    model.ori_save = model.save
    model.save = MethodType(save, model)
    return model


def _apply_pre_optimization(model, tune_cfg, sq, recover=False):
    """Retrieves sq info to absorb the scale to the layer at output channel.

    Args:
        model (QuantizationInterceptionModule): a prepared model.
        tune_cfg (OrderedDict): quantization config for ops.
        sq (TorchSmoothQuant): TorchSmoothQuant class containing sq infos.
        recover (bool, optional): whether to recover the scale. Defaults to False.
    """
    sq_max_info = {}
    if sq.record_max_info:
        sq_max_info = sq.max_value_info
    if sq_max_info:
        tsq = TorchSmoothQuant(model, None)
        alpha = tune_cfg["recipe_cfgs"]["smooth_quant_args"]["alpha"]
        for op_name, info in sq_max_info.items():
            if alpha == "auto":  # pragma: no cover
                alpha = info["alpha"]
            absorb_layer = op_name
            absorbed_layer = info["absorbed_layer"]
            input_minmax = info["input_minmax"]
            weight_max = info["weight_max"]
            if sq.weight_clip:
                weight_max = weight_max.clamp(min=1e-5)
            abs_input_max = torch.max(torch.abs(input_minmax[0]), torch.abs(input_minmax[1]))
            input_power = torch.pow(abs_input_max, alpha)
            weight_power = torch.pow(weight_max, 1 - alpha)
            scale = torch.clip(input_power / weight_power, min=1e-5)
            with torch.no_grad():
                if recover:
                    scale = 1.0 / scale
                for layer in absorbed_layer:
                    tsq._scale_layer_weight(layer, scale)
                tsq._absorb_scales(absorb_layer, 1.0 / scale)
            logger.debug(f"Current smoothquant scale of {op_name} is {scale}, alpha is {alpha}")


def _ipex_post_quant_process(model, example_inputs, use_bf16, inplace=False):
    """Converts to a jit model.

    Args:
        model (QuantizationInterceptionModule): a prepared model.
        example_inputs (tensor/tuple/dict): used to trace torch model.
        use_bf16 (bool): whether to use bf16 for mixed precision.
        inplace (bool, optional): whether to carry out model transformations in-place. Defaults to True.

    Returns:
        A converted jit model.
    """
    if (
        use_bf16 and (not os.getenv("FORCE_FP32") == "1") and (CpuInfo().bf16 or os.getenv("FORCE_BF16") == "1")
    ):  # pragma: no cover
        with torch.no_grad():
            with torch.cpu.amp.autocast():
                model = ipex.quantization.convert(model, inplace=inplace)
                try:
                    if isinstance(example_inputs, dict):
                        model = torch.jit.trace(model, example_kwarg_inputs=example_inputs)
                    else:
                        model = torch.jit.trace(model, example_inputs)
                    model = torch.jit.freeze(model.eval())
                except:
                    if isinstance(example_inputs, dict):
                        model = torch.jit.trace(
                            model, example_kwarg_inputs=example_inputs, strict=False, check_trace=False
                        )
                    else:
                        model = torch.jit.trace(model, example_inputs, strict=False)
                    model = torch.jit.freeze(model.eval())
    else:
        model = ipex.quantization.convert(model, inplace=inplace)
        with torch.no_grad():
            try:
                if isinstance(example_inputs, dict):
                    model = torch.jit.trace(model, example_kwarg_inputs=example_inputs)
                else:
                    model = torch.jit.trace(model, example_inputs)
                model = torch.jit.freeze(model.eval())
            except:  # pragma: no cover
                if isinstance(example_inputs, dict):
                    model = torch.jit.trace(model, example_kwarg_inputs=example_inputs, strict=False, check_trace=False)
                else:
                    model = torch.jit.trace(model, example_inputs, strict=False)
                model = torch.jit.freeze(model.eval())

    # After freezing, run 1 time to warm up the profiling graph executor to insert prim::profile
    # At the 2nd run, the llga pass will be triggered and the model is turned into
    # an int8 model: prim::profile will be removed and will have LlgaFusionGroup in the graph
    simple_inference(model, example_inputs, iterations=2)
    return model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

import torch

from ._quant_common.quant_config import HpDtype
from ._core.quant_dequant import QuantDequantBase
from ._core.scale_handler import update_state_dict_method, ScaleFormat
from ._core.quantized_func_wrappers import (
    init_quantized_func_wrapper_factory,
    clear_quantized_func_wrapper_factory,
    get_quantized_func_wrapper,
    OP_TYPE,
)
from .prepare_quant.prepare_model import get_world_size, get_local_rank
from .utils.logger import logger
from neural_compressor.common import options
from neural_compressor.torch.utils import (
    get_accelerator,
    is_optimum_habana_available,
    get_attr,
    set_attr,
    SaveLoadFormat,
    get_enum_from_format,
    UNIT_MAPPING,
    write_json_file,
)


MAX_FILE_SIZE = 5  # GB
cur_accelerator = get_accelerator()
tp_module_list = (
    # raw module
    "LinearLayer",
    "LinearLayerAllReduce",
    "LmHeadLinearAllreduce",
    "RowParallelLinear",
    "ColumnParallelLinear",
    "MergedColumnParallelLinear",
    "QKVParallelLinear",
    # patched module
    "PatchedLinear",
    "PatchedLinearAllReduce",
    "PatchedLmHeadLinearAllreduce",
    "PatchedColumnParallelLinear",
    "PatchedRowParallelLinear",
)


##################################### save ##################################
def remove_rank_suffix(name, local_rank, world_size):
    """Remove rank suffix from key name."""
    return name.removesuffix(f"_{local_rank}_{world_size}")


def add_rank_suffix(name, local_rank, world_size):
    """Add rank suffix to name."""
    return f"{name}_{local_rank}_{world_size}"


def save_rank_model(model, folder_prefix="", **kwargs):
    """Save state_dict for model from each rank."""
    # workaround for [SW-199005] [HQT] casted fp8 tensor cannot get data pointer
    cur_accelerator.synchronize()
    save_directory = add_rank_suffix(folder_prefix, get_local_rank(), get_world_size())
    os.makedirs(save_directory, exist_ok=True)
    safe_serialization = kwargs.get("safe_serialization", True)
    max_shard_size = kwargs.get("max_shard_size", f"{MAX_FILE_SIZE}GB")
    # save model state_dict and config.json
    model.save_pretrained(save_directory, max_shard_size=max_shard_size, safe_serialization=safe_serialization)


def find_tp_mod_list(model):
    """Get the module list that uses tensor parallel from model."""
    tp_mod_list = []
    for n, m in model.named_modules():
        if m.__class__.__name__ in tp_module_list:
            tp_mod_list.append(n)
    return tp_mod_list


def gather_state_dict(folder_prefix, file_name, tp_mod_list=[]):
    """Gather state_dict from files saved by each rank."""
    from safetensors.torch import load_file as safe_load_file

    world_size = get_world_size()

    def _is_in_list(name, tp_mod_list):
        for tp_name in tp_mod_list:
            if tp_name in name:
                return True
        return False

    gathered_state_dict = {}
    cur_state_dict = {}
    # load state_dict
    for i in range(world_size):  # TODO: assuming tp_size == world_size
        folder_name = add_rank_suffix(folder_prefix, i, world_size)
        cur_file = os.path.join(folder_name, file_name)
        cur_state_dict[i] = safe_load_file(cur_file)
    # gather state_dict
    for k, v in cur_state_dict[0].items():
        if _is_in_list(k, tp_mod_list):
            for i in range(world_size):  # TODO: assuming tp_size == world_size
                new_k = add_rank_suffix(k, i, world_size)
                gathered_state_dict[new_k] = cur_state_dict[i][k]
        else:
            gathered_state_dict[k] = cur_state_dict[0][k]
    return gathered_state_dict


def clean_rank_files(folder_prefix, file_name=None):
    """Clean files saved by each rank after gathering."""
    world_size = get_world_size()
    for i in range(world_size):  # TODO: assuming tp_size == world_size
        folder_name = add_rank_suffix(folder_prefix, i, world_size)
        if file_name is None:
            shutil.rmtree(folder_name, ignore_errors=True)
        else:
            cur_file = os.path.join(folder_name, file_name)
            os.remove(cur_file)


def parse_shard_size(size_str):
    """
    Parse a human-readable size string (e.g., '5GB', '10MB') into bytes as an integer.
    Args:
        size_str (str): The size string to parse, including a number and unit (e.g., '5GB').
    Returns:
        int: The size in bytes.
    Raises:
        ValueError: If the format of the input string is invalid.
    """
    size_str = size_str.strip().upper()

    for unit in UNIT_MAPPING:
        if size_str.endswith(unit):
            number = float(size_str[:-len(unit)].strip())
            return int(number * UNIT_MAPPING[unit])

    raise ValueError(f"Invalid size format: {size_str}")


def save_state_dict_sharded_safetensors(state_dict, save_dir, max_shard_size="5GB"):
    """Save the state_dict into safetensors files. If the model fits within max_shard_size, save as a single file.
    Otherwise, split into shards.
    Args:
        state_dict (dict): The model's state_dict.
        save_dir (str): The directory to save the files.
        max_shard_size (string): The maximum size for each file in bytes, default is 5GB.
    """
    from safetensors.torch import save_file

    os.makedirs(save_dir, exist_ok=True)

    total_model_size = sum(tensor.element_size() * tensor.numel() for tensor in state_dict.values())
    max_shard_size = parse_shard_size(max_shard_size)
    if total_model_size <= max_shard_size:
        # Save as a single file if the total size fits within max_shard_size
        single_file_path = os.path.join(save_dir, "model.safetensors")
        save_file(state_dict, single_file_path)
    else:
        # Save as multiple shards if the total size exceeds max_shard_size
        current_file_size = 0
        shard = {}
        shard_id = 1
        file_list = []

        for key, tensor in state_dict.items():
            tensor_size = tensor.element_size() * tensor.numel()

            if current_file_size + tensor_size > max_shard_size and shard:
                shard_file = os.path.join(save_dir, f"model-{shard_id:05d}-of-00000.safetensors")
                save_file(shard, shard_file)
                file_list.append(shard_file)
                shard = {}
                shard_id += 1
                current_file_size = 0

            shard[key] = tensor
            current_file_size += tensor_size

        if shard:
            shard_file = os.path.join(save_dir, f"model-{shard_id:05d}-of-00000.safetensors")
            save_file(shard, shard_file)
            file_list.append(shard_file)

        num_shards = len(file_list)
        for _, file_path in enumerate(file_list):
            new_file_name = file_path.replace("00000", f"{num_shards:05d}")
            os.rename(file_path, new_file_name)


def convert_weight_to_vllm_compatible(state_dict):
    """To convert INC fp8 model weight format to the vllm compatible format quantized by llm-compressor.
    Args:
        state_dict (dict): state_dict from INC model.
    Returns:
        The modified state_dict includes weight and scale adapted to llm-compressor.
    """
    keys_to_remove = []
    keys_to_add = {}

    for key in list(state_dict.keys()):
        if "scale_weight" in key:
            weight_scale = key.replace("scale_weight", "weight_scale")
            keys_to_add[weight_scale] = state_dict[key].contiguous().to("cpu")
            keys_to_remove.append(key)
        elif "k_cache.dequant_output.scale" in key:
            kv_scale = key.replace("k_cache.dequant_output.scale", "kv_scale")
            v_scale = key.replace("k_cache", "v_cache")
            k_scale_inv = key.replace("k_cache.dequant_output.scale", "k_cache.quant_input.scale_inv")
            v_scale_inv = key.replace("k_cache.dequant_output.scale", "v_cache.quant_input.scale_inv")

            keys_to_add[kv_scale] = max(state_dict[key], state_dict[v_scale]).contiguous().to("cpu")
            keys_to_remove.extend([key, v_scale, k_scale_inv, v_scale_inv])
        elif "scale_input" in key:
            input_scale = key.replace("scale_input", "input_scale")
            scale_input_inv = key.replace("scale_input", "quant_input.scale_inv")

            keys_to_add[input_scale] = state_dict[key].contiguous().to("cpu")
            keys_to_remove.extend([key, scale_input_inv])
        elif "proj.weight" in key:
            state_dict[key] = state_dict[key].t().contiguous().to("cpu")

    # Remove unnecessary keys
    for key in keys_to_remove:
        state_dict.pop(key, None)

    # Add new keys
    state_dict.update(keys_to_add)

    return state_dict


def generate_scheme(
    dynamic=False, observer="minmax", observer_kwargs={}, strategy="tensor", symmetric=True, type="float"
):
    """To generate the scheme requested by vllm compatible config."""
    return {
        "actorder": None,
        "block_structure": None,
        "dynamic": dynamic,
        "group_size": None,
        "num_bits": 8,
        "observer": observer,
        "observer_kwargs": observer_kwargs,
        "strategy": strategy,
        "symmetric": symmetric,
        "type": type,
    }


def convert_config_to_vllm_compatible(config):
    """Convert INC quantization_config to vllm compatible config.
    Args:
        config(dict): inc quantization config dictionary.
    """
    config_groups = {
        "group_0": {
            "input_activations": generate_scheme(
                dynamic=False, observer="minmax", observer_kwargs={}, strategy="tensor", symmetric=True, type="float"
            ),
            "output_activations": None,
            "targets": ["Linear"],
            "weights": generate_scheme(
                dynamic=False, observer="minmax", observer_kwargs={}, strategy="tensor", symmetric=True, type="float"
            ),
        }
    }
    ignore = [] if "names" not in config.blocklist else config.blocklist["names"]
    kv_cache_scheme = generate_scheme(
        dynamic=False, observer="minmax", observer_kwargs={}, strategy="tensor", symmetric=True, type="float"
    )
    quantization_config = {
        "config_groups": config_groups,
        "quant_method": "compressed-tensors",
        "format": "float-quantized",
        "ignore": ignore,
        "kv_cache_scheme": kv_cache_scheme,
    }
    return quantization_config


def check_config_for_vllm_compatible(config):
    """
    check config parameters for saving vllm compatible format.
    """
    error_message = (
        "Save format='vllm' only supports per-tensor static quantization. The allowlist only supports "
        "Linear-related types (e.g., LinearReduce, LinearLayer) and KVCache-related types (e.g., VLLMKVCache)."
    )
    assert "_PCS_" not in config.scale_method.upper(), error_message
    assert "types" in config.allowlist, error_message
    types = config.allowlist["types"]
    assert all("kvcache" in t.lower() or "linear" in t.lower() for t in types), error_message


def save_for_multi_devices(model, checkpoint_dir="saved_results", format="huggingface", **kwargs):
    """Save FP8 model for multi-devices.
    Args:
        model (torch.nn.Module): fp8 model object.
        checkpoint_dir (str, optional): path to checkpoint. Defaults to "saved_results".
        format (str, optional): defaults to 'huggingface'.
    """
    from safetensors.torch import save_file as safe_save_file
    folder_prefix = os.path.join(options.workspace, checkpoint_dir)
    save_rank_model(model, folder_prefix=folder_prefix, **kwargs)
    # Ensure all ranks have saved their model before proceeding
    if torch.distributed.is_initialized():
        torch.distributed.barrier()
    rank_directory = add_rank_suffix(folder_prefix, 0, world_size)
    files_list = find_safetensors_files(rank_directory)
    # use rank:0 process to gather checkpoint files
    if local_rank == 0:
        tp_mod_list = find_tp_mod_list(model)
        os.makedirs(checkpoint_dir, exist_ok=True)
        # get the safetensors file list from one folder
        # based on the safetensors file name to collect tensors from shard folders
        for file_name in files_list:
            gathered_state_dict = gather_state_dict(
                folder_prefix=folder_prefix, file_name=file_name, tp_mod_list=tp_mod_list
            )
            if format == SaveLoadFormat.VLLM:
                gathered_state_dict = update_to_vllm_compatible(model, gathered_state_dict)
            safe_save_file(gathered_state_dict, os.path.join(checkpoint_dir, file_name), metadata={"format": "pt"})
            clean_rank_files(folder_prefix=folder_prefix, file_name=file_name)
    if torch.distributed.is_initialized():
        torch.distributed.barrier()
    clean_rank_files(folder_prefix=folder_prefix)


def save_for_single_device(model, checkpoint_dir="saved_results", format="huggingface", **kwargs):
    """Save FP8 model for single device.

    Args:
        model (torch.nn.Module): fp8 model object.
        checkpoint_dir (str, optional): path to checkpoint. Defaults to "saved_results".
        format (str, optional): defaults to 'huggingface'.
    """
    from huggingface_hub import save_torch_model

    os.makedirs(checkpoint_dir, exist_ok=True)
    # TODO: [SW-207766] RuntimeError: "fill_empty_deterministic_" not implemented for 'Float8_e4m3fn'
    tmp_deterministic_algorithms_flag = torch.are_deterministic_algorithms_enabled()
    torch.use_deterministic_algorithms(False)
    # TODO: [SW-199005] [HQT] casted fp8 tensor cannot get data pointer
    cur_accelerator.synchronize()
    if format == SaveLoadFormat.VLLM:
        state_dict = convert_weight_to_vllm_compatible(state_dict=model.state_dict())
        save_state_dict_sharded_safetensors(state_dict, checkpoint_dir, max_shard_size=f"{MAX_FILE_SIZE}GB")
    else:
        safe_serialization = kwargs.get("safe_serialization", True)
        max_shard_size = kwargs.get("max_shard_size", f"{MAX_FILE_SIZE}GB")
        # save model state_dict and config.json
        model.save_pretrained(checkpoint_dir, max_shard_size=max_shard_size, safe_serialization=safe_serialization)
        torch.use_deterministic_algorithms(tmp_deterministic_algorithms_flag)


def save(model, checkpoint_dir="saved_results", format="huggingface", **kwargs):
    """Save FP8 model.

    Args:
        model (torch.nn.Module): fp8 model object.
        checkpoint_dir (str, optional): path to checkpoint. Defaults to "saved_results".
        format (str, optional): defaults to 'huggingface'.
    """
    world_size = get_world_size()
    local_rank = get_local_rank()
    format = get_enum_from_format(format)
    model = process_model_for_scalar_scale(model)
    if world_size > 1:
        assert torch.distributed.is_initialized(), "Distributed environment is not initialized."
    if format == SaveLoadFormat.VLLM:
        check_config_for_vllm_compatible(model.qconfig[next(iter(model.qconfig))])
    if world_size > 0:
        save_for_multi_devices(model, checkpoint_dir=checkpoint_dir, format=format, **kwargs)
    else:
        save_for_single_device(model, checkpoint_dir=checkpoint_dir, format=format, **kwargs)

    # save "quantization_config" in config.json
    if local_rank in [0, -1]:
        # Ensure those codes run on a single rank.
        configs_mapping = model.qconfig
        config_object = configs_mapping[next(iter(configs_mapping))]
        config_object.mode = "LOAD"
        config_object.world_size = world_size  # record world_size for loading
        # Flux pipeline has FrozenDict as config
        if not isinstance(model.config, dict):
            update_model_config(model, format, config_object)
            model.config.save_pretrained(checkpoint_dir)
        write_json_file(os.path.join(checkpoint_dir, "quantization_config.json"), config_object.to_dict())

        if hasattr(model, "generation_config") and model.generation_config is not None:
            model.generation_config.save_pretrained(checkpoint_dir)
    if torch.distributed.is_initialized():
        torch.distributed.barrier()


##################################### load ##################################

def load_empty_raw_model(model_name_or_path, **kwargs):
    """Initialize BF16 model with meta tensor."""
    import transformers
    from accelerate import init_empty_weights

    # Handling model objects not in AutoModelForCausalLM
    model = kwargs.get("original_model", None)
    # Handle Flux pipeline without AutoConfig
    try:
        config = transformers.AutoConfig.from_pretrained(model_name_or_path, **kwargs)
        quantization_config = config.quantization_config if hasattr(config, "quantization_config") else None
        hp_dtype = config.torch_dtype
    except:
        config, hp_dtype = model.config, torch.bfloat16
        quantization_config = kwargs.get("quantization_config", None)
        setattr(model.config, "quantization_config", quantization_config)

    if quantization_config is not None and "hp_dtype" in quantization_config:
        hp_dtype = HpDtype[quantization_config["hp_dtype"].upper()].value

    # fp8 model provided by neuralmagic.
    if (
        "quant_method" in quantization_config
        and quantization_config["quant_method"] in ["fp8", "compressed-tensors"]
    ):
        from_neuralmagic = True
        if (
            "kv_cache_scheme" in quantization_config
            and quantization_config["kv_cache_scheme"] is not None
        ):
            from_neuralmagic_with_kv = True
        else:
            from_neuralmagic_with_kv = False
    else:
        from_neuralmagic = False
        from_neuralmagic_with_kv = False

    if from_neuralmagic_with_kv:
        config.flash_attention_fp8 = True
        if is_optimum_habana_available:
            from optimum.habana.transformers.modeling_utils import adapt_transformers_to_gaudi
            adapt_transformers_to_gaudi()
        else:
            raise ValueError("Please install optimum-habana to load fp8 kv cache model.")

    if model is None:
        with init_empty_weights(include_buffers=False):
            model = transformers.AutoModelForCausalLM.from_config(config, torch_dtype=hp_dtype)
    world_size = get_world_size()
    if world_size > 1:
        import deepspeed
        from neural_compressor.torch.utils import get_non_persistent_buffers, load_non_persistent_buffers

        # TODO: [SW-199728] [DeepSpeed] Buffers initialized by model are not correct after tensor parallel
        # get_non_persistent_buffers and load_non_persistent_buffers are workarounds of [SW-199728]
        non_persistent_buffers = get_non_persistent_buffers(model)
        ds_inference_kwargs = {
            "dtype": hp_dtype,
            "tensor_parallel": {"tp_size": world_size},
        }
        model = deepspeed.init_inference(model, **ds_inference_kwargs)
        model = model.module
        load_non_persistent_buffers(model, non_persistent_buffers)
    model.to(hp_dtype)

    try:
        generation_config = transformers.GenerationConfig.from_pretrained(model_name_or_path, **kwargs)
        model.generation_config = generation_config
    except:  # Since model.generation_config is optional, relaxed exceptions can handle more situations.
        logger.warning("model.generation_config is not loaded correctly.")
    return model, from_neuralmagic, from_neuralmagic_with_kv


def find_safetensors_files(model_name_or_path, **kwargs):
    """Get safetensors file list."""
    is_local = os.path.isdir(model_name_or_path)
    if is_local:
        safetensors_files = []
        for root, dirs, files in os.walk(model_name_or_path):
            for file in files:
                if file.endswith(".safetensors"):
                    safetensors_files.append(file)
        return safetensors_files
    else:
        from transformers.modeling_utils import get_checkpoint_shard_files
        from transformers.utils import SAFE_WEIGHTS_INDEX_NAME, SAFE_WEIGHTS_NAME, cached_file

        # download online files
        resolved_archive_file = cached_file(
            model_name_or_path, SAFE_WEIGHTS_NAME, _raise_exceptions_for_missing_entries=False
        )
        if resolved_archive_file is None:
            # state_dict is sharded.
            resolved_archive_file = cached_file(model_name_or_path, SAFE_WEIGHTS_INDEX_NAME)
            # rsolved_archive_file becomes a list of files that point to the different checkpoint shards in this case.
            resolved_archive_file, sharded_metadata = get_checkpoint_shard_files(
                model_name_or_path,
                resolved_archive_file,
                **kwargs,
            )
        # for the model only with 1 model.safetensors file.
        if isinstance(resolved_archive_file, str):
            resolved_archive_file = [resolved_archive_file]
        return resolved_archive_file


def shard_state_dict(state_dict, return_all_rank=False, src_world_size=None):
    """Shard state_dict for current local_rank."""
    def get_rank_state_dict(state_dict, local_rank, world_size):
        rank_state_dict = {}
        for k, v in state_dict.items():
            if k.endswith(f"_{world_size}") and not k.endswith(f"_{local_rank}_{world_size}"):
                # only collect current rank state_dict and common state_dict (e.g., embedding)
                continue
            new_k = remove_rank_suffix(k, local_rank, world_size)
            rank_state_dict[new_k] = v
        return rank_state_dict

    if return_all_rank:
        all_rank_state_dict = {}
        for rank in range(src_world_size):
            rank_state_dict = get_rank_state_dict(state_dict, rank, src_world_size)
            all_rank_state_dict[rank] = rank_state_dict
        return all_rank_state_dict
    else:
        return get_rank_state_dict(state_dict, local_rank, world_size)


def split_rank_state_dict(model, gathered_state_dict):
    """Split state_dict for current local_rank."""
    rank_state_dict = {}
    for name, param in model.named_parameters():
        if name in gathered_state_dict:
            full_weight = gathered_state_dict[name]
            if len(param.shape) != 0 and full_weight.shape != param.shape:
                # clone to release reference to the original tensor
                if full_weight.shape[0] != param.shape[0]:
                    split_weight = split_weights(full_weight, world_size, local_rank, split_axis=0).clone()
                elif full_weight.shape[1] != param.shape[1]:
                    split_weight = split_weights(full_weight, world_size, local_rank, split_axis=1).clone()
                else:
                    split_weight = split_weights(full_weight, world_size, local_rank, split_axis=0).clone()
            else:
                split_weight = full_weight
            rank_state_dict[name] = split_weight

    return rank_state_dict


def get_inc_fp8config(model, from_neuralmagic=False, from_neuralmagic_with_kv=False):
    """Get INC FP8 Config.

    Args:
        model: empty model.
        from_neuralmagic(bool, optional): whether provided from nerualmagic modelhub.
        from_neuralmagic_with_kv(bool, optional): whether provided from nerualmagic modelhub and quantized kv_cache.

    Returns:
        INC FP8 Config.
    """
    from neural_compressor.torch.quantization import FP8Config
    if from_neuralmagic:
        if "ignore" in model.config.quantization_config.keys():
            blocklist =  {"types": [], "names": model.config.quantization_config["ignore"]}
        elif "ignored_layers" in model.config.quantization_config.keys():
            blocklist =  {"types": [], "names": model.config.quantization_config["ignored_layers"]}
        else:
            blocklist =  {"types": [], "names": ["lm_head"]}
        if "target" in model.config.quantization_config.keys():
            allowlist = {"types": model.config.quantization_config["target"], "names": []}
        else:
            if from_neuralmagic_with_kv:
                allowlist = {"types": ["Linear", "LinearLayer", "LinearAllreduce", "KVCache"], "names": []}
            else:
                allowlist = {"types": ["Linear", "LinearLayer", "LinearAllreduce"], "names": []}
        qconfig = FP8Config(mode="LOAD", allowlist=allowlist, blocklist=blocklist, scale_format="CONST")
    else:
        if hasattr(model, "qconfig") and model.qconfig is not None:
            configs_mapping = model.qconfig
            qconfig = configs_mapping[next(iter(configs_mapping))]
        else:
            qconfig = FP8Config.from_dict(model.config.quantization_config)
    return qconfig


def load(model_name_or_path, format="huggingface", device="hpu", **kwargs):
    """Load FP8 model.

    Args:
        model_name_or_path (str): model name or path.
        format (str, optional): format to load. Defaults to 'huggingface'.
        device (str, optional): device to use. Defaults to 'hpu'.
        kwargs (dict, optional): remaining dictionary of keyword arguments for loading huggingface models.

    Returns:
        FP8 model.
    """
    format = get_enum_from_format(format)
    world_size = kwargs.get("world_size", get_world_size())
    assert format == SaveLoadFormat.HUGGINGFACE, "Currently, only huggingface models are supported."
    assert device in ["hpu", "cpu"], "Currently, only hpu & cpu device is supported for FP8 model."

    from safetensors.torch import load_file as safe_load_file
    from neural_compressor.torch.algorithms.fp8_quant import prep_model

    model, from_neuralmagic, from_neuralmagic_with_kv = load_empty_raw_model(model_name_or_path, **kwargs)
    qconfig = get_inc_fp8config(model, from_neuralmagic, from_neuralmagic_with_kv)
    qconfig.save_temp_json_file()  # generate qconfig.json_file
    src_world_size = qconfig.world_size if hasattr(qconfig, "world_size") else world_size
    need_unify_weights = True if src_world_size > world_size else False
    if world_size > 1:
        assert torch.distributed.is_initialized(), "Distributed environment is not initialized."

    init_quantized_func_wrapper_factory()
    # replace modules to patched modules
    prep_model(model, qconfig.json_file)
    model = process_model_for_scalar_scale(model)
    # get the safetensors file list from one folder
    files_list = find_safetensors_files(model_name_or_path, **kwargs)  # TODO: get online files
    # based on the safetensors file name to collect tensors from shard folders
    for file_name in files_list:
        cur_file = os.path.join(model_name_or_path, file_name)
        gathered_state_dict = safe_load_file(cur_file)
        if from_neuralmagic or from_neuralmagic_with_kv:
            gathered_state_dict = convert_weight_to_inc(state_dict=gathered_state_dict)
        if src_world_size > 1:
            # only return state_dict for the current local_rank
            if from_neuralmagic or from_neuralmagic_with_kv:
                rank_state_dict = split_rank_state_dict(model, gathered_state_dict)
            else:
                rank_state_dict = shard_state_dict(gathered_state_dict, return_all_rank=need_unify_weights, src_world_size=src_world_size)
                if need_unify_weights:
                    # Assuming: src_world_size % world_size == 0
                    if src_world_size % world_size != 0:
                        raise EnvironmentError(
                            f"Original world_size: {src_world_size} must be divisible by target world_size: {world_size}."
                        )
                    rank_state_dict = get_new_rank_state_dict(rank_state_dict, model, world_size)
        else:
            rank_state_dict = gathered_state_dict
        model.load_state_dict(rank_state_dict, assign=True, strict=False)
        load_scale_params(model, rank_state_dict)  # ensure per-channel scale is loaded correctly
    clear_quantized_func_wrapper_factory()
    if hasattr(model, "tie_weights"):
        model.tie_weights()
    model = model.to(cur_accelerator.name())
    model = model.eval()
    cur_accelerator.synchronize()
    return model


def convert_weight_to_inc(state_dict):
    """To convert the vllm compatible fp8 model weight to INC format,
       one is operators' name are different, the other is to adapt weight on G2
       due to the torch.float8_e4m3fn scope [-240, 240].

    Args:
        state_dict (dict): state_dict from modelhub.

    Returns:
        state_dict includes weight and scale adapted to INC format.
    """
    import habana_frameworks.torch.utils.experimental as htexp
    on_gaudi2 = htexp._get_device_type() == htexp.synDeviceType.synDeviceGaudi2
    key_name = state_dict.keys()
    for key in list(key_name):
        if "weight_scale" in key:
            scale_weight = key.replace("weight_scale", "scale_weight")
            if on_gaudi2:
                # dequant_weight
                weight_key = key.replace("weight_scale", "weight")
                qweight = state_dict[weight_key].t().to(torch.bfloat16).to("hpu")
                scale = state_dict[key].to("hpu")
                dequant_weight = qweight * scale
                # recompute scale, qweight
                recompute_scale = scale * (torch.finfo(torch.float8_e4m3fn).max /
                                        torch.finfo(torch.float8_e4m3fnuz).max)
                qweight = torch.ops.hpu.cast_to_fp8_v2(dequant_weight, 1.0 / recompute_scale, False, False, torch.float8_e4m3fn)[0]
                state_dict[weight_key] = qweight
                state_dict[scale_weight] = recompute_scale
            else:
                state_dict[scale_weight] = state_dict[key].to("hpu")
            state_dict.pop(key)
        elif "kv_scale" in key:
            k_scale_inv = key.replace("kv_scale", "k_cache.quant_input.scale_inv")
            v_scale_inv = key.replace("kv_scale", "v_cache.quant_input.scale_inv")
            k_scale = key.replace("kv_scale", "k_cache.dequant_output.scale")
            v_scale = key.replace("kv_scale", "v_cache.dequant_output.scale")
            state_dict[k_scale_inv] = 1 / state_dict[key].to("hpu")
            state_dict[v_scale_inv] = 1 / state_dict[key].to("hpu")
            state_dict[k_scale] = state_dict[key].to("hpu")
            state_dict[v_scale] = state_dict[key].to("hpu")
            state_dict.pop(key)
        elif "input_scale" in key:
            scale_input_inv = key.replace("input_scale", "quant_input.scale_inv")
            scale_input = key.replace("input_scale", "scale_input")
            state_dict[scale_input_inv] = 1 / state_dict[key].to("hpu")
            state_dict[scale_input] = state_dict[key].to("hpu")
            state_dict.pop(key)
        elif "proj.weight" in key and not on_gaudi2:
            state_dict[key] = state_dict[key].detach().t().to("hpu")
        else:
            pass
        cur_accelerator.synchronize()
    return state_dict


def split_weights(weight, tp_size, tp_rank, split_axis=0):
    """
    Args:
        weight (torch.Tensor): weight tensor.
        tp_size (int): tensor parallel size.
        tp_rank (int): tensor parallel rank.
        split_axis (int): split by column or line, 0 or 1.
    Returns:
        torch.Tensor: split weight tensor.
    """
    split_size = weight.shape[split_axis] // tp_size
    start_idx = tp_rank * split_size
    end_idx = (tp_rank + 1) * split_size

    if len(weight.shape) == 1:
        return weight[start_idx:end_idx]
    elif split_axis == 0:
        return weight[start_idx:end_idx, :]
    elif split_axis == 1:
        return weight[:, start_idx:end_idx]
    else:
        raise ValueError("split_axis must be 0 (row) or 1 (column).")


def process_model_for_scalar_scale(model):
    from neural_compressor.torch.algorithms.fp8_quant.patched_module_base import PatchedModuleBase
    from neural_compressor.torch.algorithms.fp8_quant._core.quant_dequant import QuantDequantBase
    for mod in model.modules():
        if isinstance(mod, (PatchedModuleBase, QuantDequantBase)):
            update_state_dict_method(mod)
    return model


def update_model_config(model, format, config_object):
    """Update model config for saving."""
    # update config for tie_word_embeddings
    if model.config.tie_word_embeddings:
        from transformers.modeling_utils import _get_tied_weight_keys
        embedding_weight = model.get_input_embeddings().weight
        for key in _get_tied_weight_keys(model):
            if model.state_dict()[key].data_ptr() != embedding_weight.data_ptr():
                model.config.tie_word_embeddings = False
    # update per format
    if format == SaveLoadFormat.VLLM:
        quantization_config = convert_config_to_vllm_compatible(config_object)
        model.config.quantization_config = quantization_config
    else:
        model.config.quantization_config = config_object


def load_scale_params(model, new_scale_params):
    """
    Load scale parameters into the model. If the lengths do not match, replace the old scales with the new ones.
    
    Args:
        model (torch.nn.Module): The model to load scale parameters into.
        new_scale_params (dict): A dictionary containing the new scale parameters.
    """
    for name, param in model.named_parameters():
        if name in new_scale_params:
            new_scale = new_scale_params[name]
            old_scale = param.data
            # update per-tensor scale with per-channel scale
            if old_scale.shape != new_scale.shape:
                param.data = new_scale


def get_new_rank_state_dict(all_rank_state_dict, model, world_size=get_world_size(), local_rank=get_local_rank()):
    """Get new rank state_dict for world_size.

    Args:
        all_rank_state_dict (dict): {0: state_dict, 1: state_dict, ...} for all ranks.
        model (torch.nn.Module): A quantized model with empty weights in load mode.
        world_size (int, optional): Target world size. Defaults to world_size.
        local_rank (int, optional): Current local rank. Defaults to local_rank.
    """

    def dq_q_weight(weight, previous_scale, target_scale):
        """dequantize and quantize weight with different scales."""
        cast_to_op = get_quantized_func_wrapper(OP_TYPE.CAST_TO_FP8, ScaleFormat.CONST)
        cast_from_op = get_quantized_func_wrapper(OP_TYPE.CAST_FROM_FP8, ScaleFormat.CONST)
        hp_weight = cast_from_op(weight, previous_scale, torch.bfloat16)
        # cast_to_op should use scale_inv = 1/ scale
        fp8_weight = cast_to_op(hp_weight, 1 / target_scale, False, False, dtype=weight.dtype)
        return fp8_weight

    def reorder_dict(d):
        """Reorder dict by moving weights to the end so that scales are process first."""
        from collections import OrderedDict
        without_weight = OrderedDict((k, v) for k, v in d.items() if not k.endswith('.weight'))
        with_weight = OrderedDict((k, v) for k, v in d.items() if k.endswith('.weight'))
        reordered_dict = OrderedDict()
        reordered_dict.update(without_weight)
        reordered_dict.update(with_weight)
        return reordered_dict

    def get_src_ranks(src_world_size, dst_world_size, local_rank):
        """Get src_ranks for current local_rank. 
            e.g.
                If src_world_size=8, dst_world_size=2:
                    1. local_rank=0, src_ranks=[0, 1, 2, 3]
                    2. local_rank=1, src_ranks=[4, 5, 6, 7]
        """
        shard_size = src_world_size // dst_world_size
        rank_combination = [[] for _ in range(dst_world_size)]
        for i in range(src_world_size):
            dst_rank =  i // shard_size
            rank_combination[dst_rank].append(i)
        src_ranks = rank_combination[local_rank]
        logger.info(f"local_rank={local_rank}, src_ranks={src_ranks}")
        return src_ranks

    def adjust_local_rank_and_world_size(world_size, local_rank):
        """Adjust world_size and local_rank for single card."""
        if world_size in [0, -1]:
            world_size = 1
        if local_rank == -1:
            local_rank = 0
        return world_size, local_rank

    def check_tp_status(target_weight_name, new_rank_state_dict, params_dict, scale):
        """Check is_tp, is_per_channel, is_tp_output_dim for tensor paralleled weight."""
        assert target_weight_name in params_dict, f"Weight {target_weight_name} not found in state_dict."
        target_weight_shape = params_dict[target_weight_name].shape
        cur_weight_shape = new_rank_state_dict[target_weight_name].shape
        is_tp = True if target_weight_shape != cur_weight_shape else False
        is_per_channel = True if scale.numel() != 1 else False
        # output dim of PatchedModule is 1, LinearAllReduce is sharding input dim
        is_tp_output_dim = True if target_weight_shape[1] != cur_weight_shape[1] else False
        return is_tp, is_per_channel, is_tp_output_dim

    def update_corresponding_weight(previous_state_dict, cur_state_dict, target_weight_name, new_scale, previous_scale, cur_scale):
        """Update corresponding weight based on new_scale, previous_scale, cur_scale."""
        device = cur_accelerator.name()
        if (new_scale != previous_scale).any():
            # update previous weight
            previous_state_dict[target_weight_name] = dq_q_weight(previous_state_dict[target_weight_name], previous_scale, new_scale)
        if (new_scale != cur_scale).any():
            # update new weight
            cur_state_dict[target_weight_name] = cur_state_dict[target_weight_name].to(device)
            cur_state_dict[target_weight_name] = dq_q_weight(cur_state_dict[target_weight_name], cur_scale, new_scale)


    # Ensure world_size and local_rank are meaningful even in single card
    world_size, local_rank = adjust_local_rank_and_world_size(world_size, local_rank)
    # get source ranks for current local_rank, source ranks are used to merge state_dict
    src_ranks = get_src_ranks(len(all_rank_state_dict), world_size, local_rank)

    # merge state_dict for current local_rank
    params_dict = dict(model.named_parameters())
    device = cur_accelerator.name()
    new_rank_state_dict = {}
    # merge state_dict for current local_rank
    for rank in src_ranks:
        state_dict = all_rank_state_dict[rank]
        # reorder dict for post-process weight with all-reduced scale
        state_dict = reorder_dict(state_dict)
        for k, v in state_dict.items():
            v = v.to(device)
            if k not in new_rank_state_dict:
                new_rank_state_dict[k] = v
            else:
                module_name, param_name = k.rsplit('.', 1)
                module = get_attr(model, module_name)
                # process scales and weights based on is_per_channel, is_tp, is_tp_output_dim
                if hasattr(module, "scale_members") and param_name in module.scale_members:
                    previous_scale = new_rank_state_dict[k]
                    cur_scale = v
                    if param_name == "scale_weight":
                        target_weight_name = module_name + ".weight"
                        is_tp, is_per_channel, is_tp_output_dim = check_tp_status(
                            target_weight_name, new_rank_state_dict, params_dict, cur_scale
                        )
                    # select correct way to update scales and weights
                    if param_name == "scale_weight" and is_tp and is_per_channel and is_tp_output_dim:
                        # only cat per-channel scales and weight
                        new_rank_state_dict[k] = torch.cat([new_rank_state_dict[k], v], dim=0)
                    else:
                        # select correct scale with min/max
                        if isinstance(module, QuantDequantBase) and param_name == "scale_inv":
                            new_rank_state_dict[k] = torch.min(new_rank_state_dict[k], v)
                        else:
                            new_rank_state_dict[k] = torch.max(new_rank_state_dict[k], v)
                        # update corresponding weight based on new_scale, previous_scale, cur_scale
                        if (previous_scale != cur_scale).any() and param_name == "scale_weight":
                            new_scale = new_rank_state_dict[k]
                            update_corresponding_weight(
                                new_rank_state_dict, state_dict, target_weight_name, new_scale, previous_scale, cur_scale
                            )
                elif k in params_dict and params_dict[k].shape != new_rank_state_dict[k].shape:
                    # Detect tp dim by comparing current weight shape with the one in model, it's more robust.
                    # e.g. LinearAllReduce(input dim) and LinearLayer(output dim) are using different dims.
                    tp_dim = 0 if params_dict[k].shape[0] != new_rank_state_dict[k].shape[0] else 1
                    new_rank_state_dict[k] = torch.cat([new_rank_state_dict[k], v], dim=tp_dim)
    return new_rank_state_dict


def update_to_vllm_compatible(model, gathered_state_dict):
    """update gathered_state_dict to vllm compatible format.

    Args:
        model (model): the quantized model.
        gathered_state_dict (dict): state_dict for all ranks.
    Returns:
        gathered_state_dict (dict): vllm compatible state_dict.
    """
    # for example, tp = 2, make
    # 'model.layers.21.mlp.down_proj.weight_0_2', 'model.layers.21.mlp.down_proj.weight_1_2'
    #  --> 'model.layers.21.mlp.down_proj.weight'
    import transformers
    from accelerate import init_empty_weights
    from neural_compressor.torch.algorithms.fp8_quant import prep_model

    with init_empty_weights(include_buffers=False):
        reference_model = transformers.AutoModelForCausalLM.from_config(model.config)
    # replace modules with patched modules
    inc_config = get_inc_fp8config(model)
    inc_config.mode = "LOAD"
    inc_config.save_temp_json_file()
    prep_model(reference_model, inc_config.json_file)
    # gather weights into 1 rank
    rank_state_dict = shard_state_dict(gathered_state_dict, return_all_rank=True, src_world_size=world_size)
    rank_state_dict = get_new_rank_state_dict(rank_state_dict, reference_model, world_size=1)
    # rename param names
    gathered_state_dict = convert_weight_to_vllm_compatible(state_dict=rank_state_dict)
    return gathered_state_dict



# Copyright (c) 2023-2024 Mobiusml and Intel Corporation

# This code is based on Mobiusml's HQQ library. It has been modified
# from its original forms to simplify and adapt it for use in
# the Intel® Neural Compressor.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Notice: Copied from from https://github.com/mobiusml/hqq
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
"""Bit packing logic for HQQ."""


import numpy as np
import torch

__all__ = ["Packer"]


# Bit packing logic. format: pack/unpack_nBits_target-<uint8 or int32>
class BitPack:
    """Packing and unpacking tensors into different bit representations."""

    # 8-bit
    ################################################
    @staticmethod
    def pack_8bit_u8(W_q):
        """Packs the given tensor into 8-bit unsigned integers.

        Args:
            W_q (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The packed tensor.
        """
        return W_q.to(torch.uint8)

    @staticmethod
    def unpack_8bit_u8(W_q):
        """Unpacks the given 8-bit tensor into 8-bit unsigned integer tensor.

        Args:
            W_q (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The unpacked tensor.
        """
        return W_q

    # 4-bit
    ################################################
    @staticmethod
    def pack_4bit_u8(W_q):  # uint8 > uint8/2
        """Packs the given 4-bit tensor into 8-bit unsigned integers.

        Args:
            W_q (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The packed tensor.
        """
        W_q = W_q.to(torch.uint8)
        _step = int(len(W_q) / 2)
        return (W_q[:_step] << 4) | W_q[_step:]

    # A bit faster than the _cat version
    @staticmethod
    def unpack_4bit_u8(W_q):  # uint8/2 > uint8
        """Unpacks the given 4-bit tensor into 8-bit unsigned integers.

        Args:
            W_q (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The unpacked tensor.
        """
        _step = W_q.shape[0]
        tmp = torch.empty([2 * _step, W_q.shape[1]], dtype=torch.uint8, device=W_q.device)
        tmp[:_step] = (W_q & 0b11110000) >> 4
        tmp[_step:] = W_q & 0b00001111
        return tmp

    # 2-bit
    ################################################
    @staticmethod
    def pack_2bit_u8(W_q):  # uint8 > uint8/4
        """Packs the given 2-bit tensor into 8-bit unsigned integers.

        Args:
            W_q (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The packed tensor.
        """
        W_q = W_q.to(torch.uint8)
        _step = int(len(W_q) / 4)
        return W_q[:_step] << 6 | W_q[_step : 2 * _step] << 4 | W_q[2 * _step : 3 * _step] << 2 | W_q[3 * _step :]

    # A bit faster than the _cat version
    @staticmethod
    def unpack_2bit_u8(W_q):
        """Unpacks the tensor packed by `pack_2bit_u8`.

        Args:
            W_q (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The unpacked tensor.
        """
        _step = W_q.shape[0]
        tmp = torch.empty([4 * _step, W_q.shape[1]], dtype=torch.uint8, device=W_q.device)
        tmp[:_step] = (W_q & 0b11000000) >> 6
        tmp[_step : 2 * _step] = (W_q & 0b00110000) >> 4
        tmp[2 * _step : 3 * _step] = (W_q & 0b00001100) >> 2
        tmp[3 * _step :] = W_q & 0b00000011
        return tmp

    # 3bit
    ################################################
    @staticmethod
    def pack_3bit_32(W_q_in):
        """Packs the given 3-bit tensor into 32-bit signed integers.

        Args:
            W_q_in (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The packed tensor.
        """
        W_q = torch.zeros(
            [int(10 * np.ceil(W_q_in.shape[0] / 10.0)), W_q_in.shape[1]], device=W_q_in.device, dtype=torch.int32
        )
        W_q[: len(W_q_in)] = W_q_in
        _step = int(len(W_q) / 10)
        W_q = (
            (W_q[:_step] << 27)
            | (W_q[_step : _step * 2] << 24)
            | (W_q[_step * 2 : _step * 3] << 21)
            | (W_q[_step * 3 : _step * 4] << 18)
            | (W_q[_step * 4 : _step * 5] << 15)
            | (W_q[_step * 5 : _step * 6] << 12)
            | (W_q[_step * 6 : _step * 7] << 9)
            | (W_q[7 * _step : _step * 8] << 6)
            | (W_q[_step * 8 : _step * 9] << 3)
            | (W_q[_step * 9 :])
        )
        return W_q

    # A bit faster than _cat version
    @staticmethod
    def unpack_3bit_32(W_q):
        """Unpacks the tensor packed by `pack_3bit_32`.

        Args:
            W_q (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The unpacked tensor.
        """
        _step = W_q.shape[0]
        tmp = torch.empty([10 * _step, W_q.shape[1]], dtype=torch.uint8, device=W_q.device)
        tmp[:_step] = (W_q & 0b00111000000000000000000000000000) >> 27
        tmp[1 * _step : 2 * _step] = (W_q & 0b00000111000000000000000000000000) >> 24
        tmp[2 * _step : 3 * _step] = (W_q & 0b00000000111000000000000000000000) >> 21
        tmp[3 * _step : 4 * _step] = (W_q & 0b00000000000111000000000000000000) >> 18
        tmp[4 * _step : 5 * _step] = (W_q & 0b00000000000000111000000000000000) >> 15
        tmp[5 * _step : 6 * _step] = (W_q & 0b00000000000000000111000000000000) >> 12
        tmp[6 * _step : 7 * _step] = (W_q & 0b00000000000000000000111000000000) >> 9
        tmp[7 * _step : 8 * _step] = (W_q & 0b00000000000000000000000111000000) >> 6
        tmp[8 * _step : 9 * _step] = (W_q & 0b00000000000000000000000000111000) >> 3
        tmp[9 * _step :] = W_q & 0b00000000000000000000000000000111
        return tmp


class Packer:
    """Pack/unpack functions collection."""

    bit_to_packing = {8: "8bit_u8", 4: "4bit_u8", 3: "3bit_32", 2: "2bit_u8"}

    pack_fn_mapping = {
        "8bit_u8": BitPack.pack_8bit_u8,
        "4bit_u8": BitPack.pack_4bit_u8,
        "3bit_32": BitPack.pack_3bit_32,
        "2bit_u8": BitPack.pack_2bit_u8,
    }

    unpack_fn_mapping = {
        "8bit_u8": BitPack.unpack_8bit_u8,
        "4bit_u8": BitPack.unpack_4bit_u8,
        "3bit_32": BitPack.unpack_3bit_32,
        "2bit_u8": BitPack.unpack_2bit_u8,
    }

    @staticmethod
    def get_pack_fn(nbits: int):
        """Get the pack function for the specified number of bits.

        Args:
            nbits (int): The number of bits.

        Returns:
            function: The pack function for the specified number of bits.
        """
        return Packer.pack_fn_mapping[Packer.bit_to_packing[nbits]]

    @staticmethod
    def get_unpack_fn(nbits: int):
        """Get the unpack function for the specified number of bits.

        Args:
            nbits (int): The number of bits.

        Returns:
            function: The unpack function for the specified number of bits.
        """
        return Packer.unpack_fn_mapping[Packer.bit_to_packing[nbits]]



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import abstractmethod

import torch
from neural_compressor.torch.algorithms.fp8_quant._quant_common.quant_config import get_hqt_config, is_supported_dynamic_op
from .scale_method_factory import ScaleMethodFactory, QuantTensorName, ScaleValueType
from ..common import ModuleConfig, QuantTensorType
from ..quant_dequant import DequantOutput, QuantDequant, QuantDequantNone, QuantInput, QuantDynamicInput
from ...utils.logger import logger
from neural_compressor.torch.algorithms.fp8_quant._core.common import dequant_original_fp8_weight_if_needed
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator
cur_device = auto_detect_accelerator().current_device_name()


class BaseOpQuantizer:

    def __init__(self, config, mod, measurement, params, mod_type_str):
        hqt_config = get_hqt_config(mod).cfg
        module_type = hqt_config["mod_dict"][mod_type_str]
        self.scales_method_factory = ScaleMethodFactory(config, params, mod, module_type)
        self.mod = mod
        self.params = params
        self.measurement = measurement
        self.inputs_scales_creators = []
        self.output_scales_creators = []
        self.params_scales_creators = []
        self.is_dynamic = hqt_config["dynamic_quantization"] and is_supported_dynamic_op(mod_type_str)

        logger.debug("%s %s", self.__class__.__name__, self.__dict__)

    def get_module_configuration(self):
        scale_format = get_hqt_config(self.mod).cfg["scale_format"]
        use_qdq = get_hqt_config(self.mod).cfg["use_qdq"]
        fake_quant = get_hqt_config(self.mod).cfg["fake_quant"]
        lp_dtype = self.params["lp_dtype"]
        hp_dtype = self.params["hp_dtype"]
        return scale_format, use_qdq, fake_quant, lp_dtype, hp_dtype

    @abstractmethod
    def get_scales_module_config(self) -> ModuleConfig:
        raise NotImplementedError("`get_scales_module_config` function is not implemented")

    @abstractmethod
    def scales_module_config_to_q_and_dq(self, module) -> ModuleConfig:
        raise NotImplementedError("`scales_module_config_to_q_and_dq` function is not implemented")

    def init_scales_from_module_config(self, module):
        for idx, input in enumerate(module.inputs):
            if self.inputs_scales_creators[idx].scale is None:
                self.inputs_scales_creators[idx].scale = input
        for idx, output in enumerate(module.outputs):
            if self.output_scales_creators[idx].scale is None:
                self.output_scales_creators[idx].scale = output

    def calc_input_scales(self, num_of_inputs):
        input_scales = []
        for i in range(num_of_inputs):
            input_measurement = self.measurement.inputs[i] if self.measurement is not None else []
            input_scale = None
            if not self.is_dynamic:
                input_scale = self.inputs_scales_creators[i].calc_scales(
                    input_measurement, QuantTensorType.MEASUREMENTS
                )
            input_scales.append(input_scale)
        return input_scales

    def calc_output_scales(self):
        output_measurement = self.measurement.outputs[0] if self.measurement is not None else []
        output_scales = None
        if not self.is_dynamic:
            output_scales = self.output_scales_creators[0].calc_scales(output_measurement, QuantTensorType.MEASUREMENTS)
        return (output_scales,)

    def init_input_config(self, scales_inv, lp_dtype, hp_dtype, scale_format, use_qdq, fake_quant):
        if use_qdq or fake_quant:
            input_config = [
                QuantDequant(s_inv, lp_dtype, hp_dtype, scale_format=scale_format, use_qdq=use_qdq)
                for s_inv in scales_inv
            ]
        else:
            input_config = []
            for input_scales_creator, s_inv in zip(self.inputs_scales_creators, scales_inv):
                if self.is_dynamic:
                    input_config.append(
                        QuantDynamicInput(input_scales_creator, lp_dtype, hp_dtype, scale_format=scale_format)
                    )
                else:
                    input_config.append(QuantInput(s_inv, lp_dtype, hp_dtype, scale_format=scale_format))
        return input_config


class LinearOpQuantizer(BaseOpQuantizer):

    def __init__(self, config, mod, measurement, params, mod_type_str):
        super().__init__(config, mod, measurement, params, mod_type_str)
        hqt_config = get_hqt_config(mod).cfg
        module_type = hqt_config["mod_dict"][mod_type_str]
        if module_type == "row_parallel_linear" and get_hqt_config(mod).cfg["row_parallel_linear_allreduce_quantization"]:
            self.scales_method_factory.output_scale_method_config.backoff = 1.0
        self.inputs_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.INPUT, self.is_dynamic))
        self.weight_och_scale_calc = self.scales_method_factory.get_scale_method(QuantTensorName.WEIGHT_OUT_CH)
        self.weight_ich_scale_calc = self.scales_method_factory.get_scale_method(QuantTensorName.WEIGHT_IN_CH)
        self.output_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.OUTPUT, self.is_dynamic))

    def get_scales_module_config(self):
        input_scales = self.calc_input_scales(num_of_inputs=1)
        output_measurement = self.measurement.outputs[0] if self.measurement is not None else []
        rescaled_weight = self.mod.weight if hasattr(self.mod, 'weight') else None
        if rescaled_weight is not None:
            rescaled_weight = dequant_original_fp8_weight_if_needed(self.mod, rescaled_weight)
        if self.scales_method_factory.scale_method_config_map[QuantTensorName.WEIGHT_IN_CH].scale_value_type != ScaleValueType.DUMMY_SCALES:
            # Calculating weight in hpu to support scale calculation CGUID torch.ops.hpu.calculate_scale_for_cast
            rescaled_weight = rescaled_weight.to(cur_device)
        if self.weight_ich_scale_calc is not None:
            weight_scales_in_ch = self.weight_ich_scale_calc.calc_scales(input_scales[0], QuantTensorType.CONST)
            rescaled_weight = torch.div(rescaled_weight, weight_scales_in_ch.reshape([1, -1]))
        weights_scales_out_ch = self.weight_och_scale_calc.calc_scales(rescaled_weight, QuantTensorType.CONST)

        params_config = (
            {"weight": weights_scales_out_ch}
            if (self.weight_ich_scale_calc is None)
            else {"weight": {0: weights_scales_out_ch, 1: weight_scales_in_ch}}
        )
        output_scales = None
        if not self.is_dynamic:
            output_scales = self.output_scales_creators[0].calc_scales(
                output_measurement, QuantTensorType.MEASUREMENTS, input0=weights_scales_out_ch, input1=input_scales[0]
            )
        return ModuleConfig(
            input_scales,
            (output_scales,),
            params_config,
        )

    def init_weight_config(self, scales, scales_inv, lp_dtype, hp_dtype, scale_format, use_qdq, fake_quant):
        if use_qdq:
            # to ensure the weights to be loaded to the device in fp8
            weight_config = [
                QuantInput(scales_inv, lp_dtype, hp_dtype, scale_format=scale_format, use_qdq=use_qdq),
                DequantOutput(scales, lp_dtype, hp_dtype, scale_format=scale_format, use_qdq=use_qdq),
            ]
        elif fake_quant:
            weight_config = [QuantDequant(scales_inv, lp_dtype, hp_dtype, scale_format=scale_format)]
        else:
            weight_config = [QuantInput(scales_inv, lp_dtype, hp_dtype, scale_format=scale_format)]
        return weight_config

    def init_weights_from_module(self, params_config):
        if isinstance(params_config, dict):
            self.weight_och_scale_calc.scale = params_config[0]
            self.weight_ich_scale_calc.scale = params_config[1]
        else:
            self.weight_och_scale_calc.scale = params_config

    def get_output_config(self, lp_dtype, hp_dtype, scale_format):
        output_config = [QuantDequantNone(lp_dtype, hp_dtype, scale_format=scale_format)]
        return output_config

    def scales_module_config_to_q_and_dq(self, module):
        self.init_scales_from_module_config(module)
        self.init_weights_from_module(module.params["weight"])
        scale_format, use_qdq, fake_quant, lp_dtype, hp_dtype = self.get_module_configuration()
        input_config = super().init_input_config(
            (self.inputs_scales_creators[0].calc_invert_scales(),),
            lp_dtype,
            hp_dtype,
            scale_format,
            use_qdq,
            fake_quant,
        )
        # outputs as bf16, and descaled in gemm under PatchedLinear, so no need to work here
        output_config = self.get_output_config(lp_dtype, hp_dtype, scale_format=scale_format)
        weight_config = self.init_weight_config(
            self.weight_och_scale_calc.scale,
            self.weight_och_scale_calc.calc_invert_scales(),
            lp_dtype,
            hp_dtype,
            scale_format,
            use_qdq,
            fake_quant,
        )
        params_config = {"weight": weight_config}
        if hasattr(self.mod, "bias") and (getattr(self.mod, "bias") is not None):
            # In PatchedLinear the bias is added to the output of gemm.
            # The output is expected to be descaled and in bf16, so we don't need to touch the bias.
            bias_config = [QuantDequantNone(lp_dtype, hp_dtype)]
            params_config.update({"bias": bias_config})
        return ModuleConfig(input_config, output_config, params_config)

class RowParallelLinearOpQuantizer(LinearOpQuantizer):
    def __init__(self, config, mod, measurement, params, mod_type_str):
        super().__init__(config, mod, measurement, params, mod_type_str)
        self.allreduce_quantization_enabled = get_hqt_config(mod).cfg["row_parallel_linear_allreduce_quantization"]
        if self.allreduce_quantization_enabled:
            self.output_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.OUTPUT))

    def init_scales_from_module_config(self, module):
        if not self.allreduce_quantization_enabled:
            return super().init_scales_from_module_config(module)
        for idx, input in enumerate(module.inputs):
            if self.inputs_scales_creators[idx].scale is None:
                self.inputs_scales_creators[idx].scale = input
        if self.output_scales_creators[0].scale is None:
            self.output_scales_creators[0].scale = module.outputs[0]
        if self.allreduce_quantization_enabled and self.output_scales_creators[1].scale is None:
                self.output_scales_creators[1].scale = module.outputs[1]

    def get_scales_module_config(self):
        if not self.allreduce_quantization_enabled:
            return super().get_scales_module_config()
        module_config = super().get_scales_module_config()
        output_measurement = self.measurement.outputs[1] if self.measurement is not None else []
        output_scales = self.output_scales_creators[1].calc_scales(output_measurement, QuantTensorType.MEASUREMENTS)
        module_config.outputs = (module_config.outputs[0], output_scales,)
        return module_config

    def get_output_config(self, lp_dtype, hp_dtype, scale_format):
        if not self.allreduce_quantization_enabled:
            return super().get_output_config(lp_dtype, hp_dtype, scale_format)
        scale_0 = self.output_scales_creators[0].scale
        inv_scale_0 = self.output_scales_creators[0].calc_invert_scales()
        output_config_dq_scatter_output = DequantOutput(scale_0, lp_dtype, hp_dtype, scale_format=scale_format)
        output_config_q_scatter_input = QuantInput(inv_scale_0, lp_dtype, hp_dtype, scale_format=scale_format)
        output_config = [output_config_dq_scatter_output,
                         output_config_q_scatter_input]
        inv_scale_1 = self.output_scales_creators[1].calc_invert_scales()
        scale_1 = self.output_scales_creators[1].scale
        output_config_q_gather_input = QuantInput(inv_scale_1, lp_dtype, hp_dtype, scale_format=scale_format)
        output_config_dq_gather_output = DequantOutput(scale_1, lp_dtype, hp_dtype, scale_format=scale_format)
        output_config.extend([output_config_q_gather_input, output_config_dq_gather_output])
        return output_config

class MatmulOpQuantizer(BaseOpQuantizer):

    def __init__(self, config, mod, measurement, params, mod_type_str):
        super().__init__(config, mod, measurement, params, mod_type_str)
        self.inputs_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.INPUT))
        self.inputs_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.INPUT))
        self.output_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.OUTPUT))

    def get_scales_module_config(self):
        input_scales = self.calc_input_scales(num_of_inputs=2)

        output_scales = input_scales[0] * input_scales[1]
        return ModuleConfig(input_scales, (output_scales,), {})

    def scales_module_config_to_q_and_dq(self, module):
        self.init_scales_from_module_config(module)
        scale_format, use_qdq, fake_quant, lp_dtype, hp_dtype = super().get_module_configuration()
        input_config = super().init_input_config(
            (self.inputs_scales_creators[0].calc_invert_scales(), self.inputs_scales_creators[1].calc_invert_scales()),
            lp_dtype,
            hp_dtype,
            scale_format,
            use_qdq,
            fake_quant,
        )

        # 4bit->8bit inputs, no need to quant
        if hasattr(self.mod, "no_input_quant"):
            input_config[1] = QuantDequantNone(lp_dtype, hp_dtype, scale_format=scale_format)

        # outputs as bf16, and descaled in gemm under PatchedLinear, so no need to work here
        output_config = [QuantDequantNone(lp_dtype, hp_dtype, scale_format=scale_format)]
        return ModuleConfig(input_config, output_config)


class SoftmaxOpQuantizer(BaseOpQuantizer):

    def __init__(self, config, mod, measurement, params, mod_type_str):
        super().__init__(config, mod, measurement, params, mod_type_str)
        self.output_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.OUTPUT))

    def get_scales_module_config(self):
        output_scales = self.calc_output_scales()

        return ModuleConfig((), output_scales)

    def scales_module_config_to_q_and_dq(self, module):
        self.init_scales_from_module_config(module)
        scale_format, use_qdq, fake_quant, lp_dtype, hp_dtype = super().get_module_configuration()
        output_config = [
            DequantOutput(self.output_scales_creators[0].scale, lp_dtype, hp_dtype, scale_format=scale_format)
        ]
        return ModuleConfig([], output_config, {})


class FsdpaOpQuantizer(BaseOpQuantizer):

    def __init__(self, config, mod, measurement, params, mod_type_str):
        super().__init__(config, mod, measurement, params, mod_type_str)
        self.num_of_inputs = 4
        self.inputs_scales_creators = [
            self.scales_method_factory.get_scale_method(QuantTensorName.INPUT) for i in range(self.num_of_inputs)
        ]
        self.output_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.OUTPUT))

    def get_scales_module_config(self):
        # 3 inputs calcs from input measurement
        input_scales = self.calc_input_scales(num_of_inputs=self.num_of_inputs - 1)
        # one input calcs from output measurement
        output1_measurement = self.measurement.outputs[1] if self.measurement is not None else []
        input_scales.append(
            self.inputs_scales_creators[self.num_of_inputs - 1].calc_scales(
                output1_measurement, QuantTensorType.MEASUREMENTS
            )
        )
        output_scales = self.calc_output_scales()
        return ModuleConfig(input_scales, output_scales, {})

    def scales_module_config_to_q_and_dq(self, module):
        self.init_scales_from_module_config(module)
        scale_format, use_qdq, fake_quant, lp_dtype, hp_dtype = super().get_module_configuration()
        input_scales_inv = [
            self.inputs_scales_creators[i].calc_invert_scales() for i in range(len(self.inputs_scales_creators))
        ]
        input_config = super().init_input_config(
            input_scales_inv, lp_dtype, hp_dtype, scale_format, use_qdq, fake_quant
        )
        output_config = [
            DequantOutput(self.output_scales_creators[0].scale, lp_dtype, hp_dtype, scale_format=scale_format)
        ]
        return ModuleConfig(input_config, output_config, {})


class KVCacheOpQuantizer(BaseOpQuantizer):

    def __init__(self, config, mod, measurement, params, mod_type_str):
        super().__init__(config, mod, measurement, params, mod_type_str)
        self.inputs_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.INPUT))
        self.output_scales_creators.append(self.inputs_scales_creators[0])

    # TODO: Remove after implementing lp_dtype in OHF.
    def init_input_config(self, scales_inv, lp_dtype, hp_dtype, scale_format, use_qdq, fake_quant):
        input_config = super().init_input_config(scales_inv, lp_dtype, hp_dtype, scale_format, False, fake_quant)
        if use_qdq:
            input_config.extend([
                QuantDequant(s_inv, lp_dtype, hp_dtype, scale_format=scale_format, use_qdq=use_qdq)
                for s_inv in scales_inv
            ])
        return input_config

    def get_scales_module_config(self):
        input_scales = self.calc_input_scales(num_of_inputs=1)
        self.output_scales_creators[0].scale = self.inputs_scales_creators[0].scale
        output_scales = [self.output_scales_creators[0].scale]
        return ModuleConfig(input_scales, output_scales, {})

    def scales_module_config_to_q_and_dq(self, module):
        self.init_scales_from_module_config(module)
        scale_format, use_qdq, fake_quant, lp_dtype, hp_dtype = super().get_module_configuration()
        input_scales_inv = [
            self.inputs_scales_creators[i].calc_invert_scales() for i in range(len(self.inputs_scales_creators))
        ]
        # TODO: After implementing lp_dtype in OHF can call:
        # `super().init_input_config(scales_inv, lp_dtype, hp_dtype, scale_format, False, fake_quant)`
        input_config = self.init_input_config(
            input_scales_inv, lp_dtype, hp_dtype, scale_format, use_qdq, fake_quant
        )
        output_config = [
            DequantOutput(self.output_scales_creators[0].scale, lp_dtype, hp_dtype, scale_format=scale_format, use_qdq=False)
        ]
        return ModuleConfig(input_config, output_config)


class DynamicMoeOpQuantizer(BaseOpQuantizer):

    def __init__(self, config, mod, measurement, params, mod_type_str):
        super().__init__(config, mod, measurement, params, mod_type_str)
        num_of_inputs = len(self.measurement.inputs) if self.measurement is not None else 1
        if hasattr(self.mod, "local_num_experts"):
            num_of_experts = self.mod.local_num_experts
        elif hasattr(self.mod, "num_experts"):
            num_of_experts = self.mod.num_experts
        else:
            num_of_experts = 8
        
        self.inputs_scales_creators = [
            self.scales_method_factory.get_scale_method(QuantTensorName.INPUT, is_dynamic=self.is_dynamic)
            for i in range(num_of_inputs + num_of_experts)
        ]
        self.output_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.OUTPUT))

    def get_scales_module_config(self):
        num_of_inputs = len(self.measurement.inputs) if self.measurement is not None else 1
        if hasattr(self.mod, "local_num_experts"):
            num_of_experts = self.mod.local_num_experts
        elif hasattr(self.mod, "num_experts"):
            num_of_experts = self.mod.num_experts
        else:
            num_of_experts = 8
        input_scales = self.calc_input_scales(num_of_inputs=num_of_inputs)
        for i in range(num_of_experts):
            output_measurement = self.measurement.outputs[i + 1] if self.measurement is not None else []
            input_scales.append(
                self.inputs_scales_creators[num_of_inputs + i].calc_scales(
                    output_measurement, QuantTensorType.MEASUREMENTS
                )
            )
        output_scales = self.calc_output_scales()
        return ModuleConfig(input_scales, output_scales, {})

    def scales_module_config_to_q_and_dq(self, module):
        self.init_scales_from_module_config(module)
        scale_format, use_qdq, fake_quant, lp_dtype, hp_dtype = super().get_module_configuration()
        input_scales_inv = [
            self.inputs_scales_creators[i].calc_invert_scales() for i in range(len(self.inputs_scales_creators))
        ]
        input_config = super().init_input_config(
            input_scales_inv, lp_dtype, hp_dtype, scale_format, use_qdq, fake_quant
        )
        output_config = [QuantDequantNone(lp_dtype, hp_dtype, scale_format=scale_format)]
        return ModuleConfig(input_config, output_config)

    

class EmbeddingOpQuantizer(BaseOpQuantizer):

    def __init__(self, config, mod, measurement, params, module_type):
        super().__init__(config, mod, measurement, params, module_type)
        self.inputs_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.INPUT))
        self.weight_och_scale_calc = self.scales_method_factory.get_scale_method(QuantTensorName.WEIGHT_OUT_CH)
        self.weight_ich_scale_calc = self.scales_method_factory.get_scale_method(QuantTensorName.WEIGHT_IN_CH)
        self.output_scales_creators.append(self.scales_method_factory.get_scale_method(QuantTensorName.OUTPUT))

    def get_scales_module_config(self):
        weight = self.mod.weight if hasattr(self.mod, 'weight') else None
        input_scales = self.calc_input_scales(num_of_inputs=1)

        if self.weight_ich_scale_calc is not None:
            weight_scales_in_ch = self.weight_ich_scale_calc.calc_scales(input_scales[0], QuantTensorType.CONST)
            weight = torch.div(weight, weight_scales_in_ch.reshape([1, -1]))
        weights_scales_out_ch = self.weight_och_scale_calc.calc_scales(weight, QuantTensorType.CONST)

        params_config = (
            {"weight": weights_scales_out_ch}
            if (self.weight_ich_scale_calc is None)
            else {"weight": {0: weights_scales_out_ch, 1: weight_scales_in_ch}}
        )
        return ModuleConfig(
            (),
            (),
            params_config,
        )

    def init_weight_config(self, scales, scales_inv, lp_dtype, hp_dtype, scale_format, use_qdq, fake_quant):
        if use_qdq:
            # to ensure the weights to be loaded to the device in fp8
            weight_config = [
                QuantInput(scales_inv, lp_dtype, hp_dtype, scale_format=scale_format, use_qdq=use_qdq),
                DequantOutput(scales, lp_dtype, hp_dtype, scale_format=scale_format, use_qdq=use_qdq),
            ]
        else:
            raise ValueError("For FP8 quantization, {} only supports QDQ mode now!".format(self.mod.__class__.__name__))
        return weight_config

    def init_weights_from_module(self, params_config):
        if isinstance(params_config, dict):
            self.weight_och_scale_calc.scale = params_config[0]
            self.weight_ich_scale_calc.scale = params_config[1]
        else:
            self.weight_och_scale_calc.scale = params_config

    def scales_module_config_to_q_and_dq(self, module):
        self.init_scales_from_module_config(module)
        self.init_weights_from_module(module.params["weight"])
        scale_format, use_qdq, fake_quant, lp_dtype, hp_dtype = self.get_module_configuration()
        weight_config = self.init_weight_config(
            self.weight_och_scale_calc.scale,
            self.weight_och_scale_calc.calc_invert_scales(),
            lp_dtype,
            hp_dtype,
            scale_format,
            use_qdq,
            fake_quant,
        )
        params_config = {"weight": weight_config}
        return ModuleConfig([], [], params_config)


ops_quantizer_map = {"linear": LinearOpQuantizer,
                      "matmul": MatmulOpQuantizer,
                      "fused_sdpa": FsdpaOpQuantizer,
                      "softmax": SoftmaxOpQuantizer,
                      "kv_cache": KVCacheOpQuantizer,
                      "dynamic_moe": DynamicMoeOpQuantizer,
                      "row_parallel_linear": RowParallelLinearOpQuantizer,
                      "embedding": EmbeddingOpQuantizer,
                     }

def get_op_quantizer(config, mod, measurement, params, mod_type_str):
    hqt_config = get_hqt_config(mod).cfg
    module_type = hqt_config["mod_dict"][mod_type_str]
    return ops_quantizer_map[module_type](config, mod, measurement, params, mod_type_str)



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Export quantized hf model to compatible formats."""

import tempfile
import warnings
from pathlib import Path
from typing import Any

import torch
import torch.nn as nn


def _export_quantized_weight(sub_module: nn.Module, quantization_format: str = None, weight_name: str = "weight"):
    """For the given weight attr of the sub_module, export the quantization info of it.

    The export includes converting weight tensor to correct quantized values and quantized dtype,
    and registering scaling factors.
    """
    if quantization_format is None:
        return

    weight: nn.Parameter = getattr(sub_module, weight_name)
    weight_quantizer = getattr(sub_module, "weight_quantizer")

    qdq_weight, scale = weight_quantizer._fake_quantize(weight)

    # TODO: support more scale dtype when there are other quantization format except mxfp8/mxfp4
    quantized_weight, e8m0_scale = weight_quantizer.weight_pack(qdq_weight, scale)

    sub_module.register_buffer("weight_scale", e8m0_scale)

    if quantization_format == "MXFP8":
        setattr(sub_module, weight_name, nn.Parameter(quantized_weight, requires_grad=False))

    if quantization_format == "MXFP4":
        delattr(sub_module, weight_name)
        # name aligned for vllm emulation
        sub_module.register_buffer("weight_packed", quantized_weight)


def _export_hf_checkpoint(model: nn.Module, scheme: str | None = None) -> tuple[dict[str, Any], dict[str, Any]]:
    """Exports the torch model to the packed checkpoint with original HF naming.

    The packed checkpoint will be consumed by the TensorRT-LLM unified converter.

    Args:
        model: the torch model.
        dtype: the weights data type to export the unquantized layers or the default model data type if None.

    Returns:
        post_state_dict: Dict containing quantized weights
        quant_config: config information to export hf_quant_cfg.json
    """
    # Create a model layer pool
    # If `model.model` exists use that, otherwise use `model` itself, e.g., Nemotron-H
    root = getattr(model, "model", model)
    # If that has a `.layers`, use it, otherwise fall back to the object itself
    root = getattr(root, "layers", root)
    layer_pool = {f"model.layers.{name}": sub_module for name, sub_module in root.named_modules()}

    from ..algorithms.qat.quant_utils import get_quant_config, get_quantization_format, is_quantlinear

    # compressored config
    quant_config = get_quant_config(scheme=scheme)

    for name, sub_module in layer_pool.items():
        quantization_format = get_quantization_format(sub_module)
        if quantization_format is not None:
            if is_quantlinear(sub_module):
                _export_quantized_weight(sub_module, quantization_format)

    quantized_state_dict = model.state_dict()

    return quantized_state_dict, quant_config


def export_hf2compressored_model(model: nn.Module, export_dir: Path | str = tempfile.gettempdir(), scheme: str = None):
    """Exports the torch model to the packed checkpoint with original HF naming.

    The packed checkpoint will be consumed by the VLLM.
    """
    export_dir = Path(export_dir)
    export_dir.mkdir(parents=True, exist_ok=True)

    try:
        _, quant_config = _export_hf_checkpoint(model, scheme)
        model.save_pretrained(export_dir)
        model.config.quantization_config = quant_config
        model.config.save_pretrained(export_dir)

    except Exception as e:
        warnings.warn("Cannot export model and config, the state can be saved with torch.save for further inspection.")
        raise e



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import tempfile
from collections import namedtuple

import torch

from neural_compressor.torch.algorithms.fp8_quant._quant_common.quant_config import Fp8cfg
from neural_compressor.torch.algorithms.fp8_quant.prepare_quant.prepare_model import finish_measurements


def save_calib_result(model):
    if hasattr(model, "__hqt_config__") and isinstance(model.__hqt_config__, Fp8cfg):
        # TODO SW-184714 modify hqt notation to inc notation once code is ported
        finish_measurements(model)
    else:
        raise NotImplementedError("Saving calibration results currently supported only in HPU.")


def update_mode(config_path, measure_step=False, quant_step=False):
    with open(config_path, "r") as file:
        config = json.load(file)

    if (measure_step and config.get("mode") == "MEASURE") or (quant_step and config.get("mode") == "QUANTIZE"):
        return config_path
    else:
        if measure_step:
            config["mode"] = "MEASURE"
        if quant_step:
            config["mode"] = "QUANTIZE"

        temp_file = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
        temp_file_path = temp_file.name

        with open(temp_file_path, "w") as temp_file:
            json.dump(config, temp_file)

        return temp_file_path


def generate_model_info(model):
    mod_inst_info = namedtuple("ModInstInfo", ["name", "parent"])
    parent_child_mod_dict = {}

    def create_mod_info_recursion(parent):
        for name, mod in parent.named_children():
            parent_child_mod_dict[mod] = mod_inst_info(name=name, parent=parent)
            create_mod_info_recursion(mod)

    create_mod_info_recursion(model)
    return parent_child_mod_dict


def get_patched_mod_list():
    from ._core.patching_common import mod_default_dict

    patched_mod_list = []
    for patched_mod in mod_default_dict.values():
        patched_mod_list.append(patched_mod.patched_module.__name__)
    return patched_mod_list


def restore_patched_module(patched_model):
    patched_mod_list = get_patched_mod_list()
    parent_child_mod_dict = generate_model_info(patched_model)
    # restore the original module
    with torch.no_grad():
        for name, patched_mod in patched_model.named_modules():
            patched_mod_type_str = patched_mod.__class__.__name__
            if patched_mod_type_str in patched_mod_list:
                parent = parent_child_mod_dict[patched_mod].parent
                name = parent_child_mod_dict[patched_mod].name
                origin_mod = patched_mod.orig_mod
                setattr(parent, name, origin_mod)


def with_patched_module(model):
    patched_mod_list = get_patched_mod_list()

    for name, mod in model.named_modules():
        mod_type = mod.__class__.__name__
        if mod_type in patched_mod_list:
            return True
    return False



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..quantized_func_wrapper import QuantizedFuncWrapperBase, OP_TYPE, QuantizedFuncWrapperFactory
from neural_compressor.torch.algorithms.fp8_quant._quant_common.quant_config import ScaleFormat

import torch

from abc import ABCMeta



class QuantizedXPUFuncWrapperBase(QuantizedFuncWrapperBase, metaclass=ABCMeta):
    """
    Placeholder for base class for XPU quantized func wrapper.
    """
    def __init__(self, scale_format, is_dynamic=False):
        self._quantized_func_ = self.get_default_quantized_func()

class QuantizedXPUMatmul(QuantizedXPUFuncWrapperBase):
    def get_default_quantized_func(self):
        # TODO FSW-11669 modify func once fp8_gemm_v2 is implemented
        return torch.ops.torch_ipex.fp8_gemm

    # only specific arguments are defined, to avoid having all other arguments defined in each call in patched modules.
    def __call__(self, input, other, out=None, out_dtype=torch.bfloat16, scale_input_inv=None, scale_other_inv=None):
        # TODO FSW-11669 modify call arguments once fp8_gemm_v2 is implemented
        # Current ipex ops fp8_gemm API is misaligned to hpu ops API.
        # below args are according to ipex ops to allow basic unit testing, but won't support integration in
        # INC patched modules.
        return self._quantized_func_(input,
                                     torch.bfloat16,
                                     0,
                                     other,
                                     torch.bfloat16,
                                     1,
                                     None, # bias
                                     scale_input_inv,
                                     scale_other_inv
                                    )


class QuantizedXPUCastToFP8Base(QuantizedXPUFuncWrapperBase):
    def get_default_quantized_func(self):
        return torch.ops.torch_ipex.cast_to_fp8

    def __call__(self, *args, **kwargs):
        return self._quantized_func_(*args, **kwargs)[0]

class QuantizedXPUCastFromFP8Base(QuantizedXPUFuncWrapperBase):
    def get_default_quantized_func(self):
        return torch.ops.torch_ipex.cast_from_fp8


class QuantizedXPUQuant(QuantizedXPUFuncWrapperBase):

    def get_default_quantized_func(self):
        return torch.ops.quantized_decomposed.quantize_per_tensor

    def get_scalar_quantized_func(self):
        return self.get_default_quantized_func()

    def __call__(self, input, scale, zero_point=None, axis=0, quant_min=None, quant_max=None, dtype=torch.float8_e4m3fn):
        return self._quantized_func_(input, scale, zero_point, quant_min, quant_max, dtype=dtype)


class QuantizedXPUDeQuant(QuantizedXPUFuncWrapperBase):

    def get_default_quantized_func(self):
        return torch.ops.quantized_decomposed.dequantize_per_tensor

    def get_scalar_quantized_func(self):
        return self.get_default_quantized_func()

    def __call__(self, input, scale, zero_point=None, axis=0, quant_min=None, quant_max=None, dtype=torch.float8_e4m3fn, out_dtype=torch.bfloat16):
        return self._quantized_func_(input, scale, zero_point, quant_min, quant_max, dtype=dtype, out_dtype=out_dtype)


class QuantizedXPUQuantPC(QuantizedXPUFuncWrapperBase):

    def get_default_quantized_func(self):
        return torch.ops.quantized_decomposed.quantize_per_channel

    def get_scalar_quantized_func(self):
        return self.get_default_quantized_func()

    def __call__(self, input, scale, zero_point=None, axis=0, quant_min=None, quant_max=None, dtype=torch.float8_e4m3fn):
        return self._quantized_func_(input, scale, zero_point, axis, quant_min, quant_max, dtype=dtype)


class QuantizedXPUDeQuantPC(QuantizedXPUFuncWrapperBase):

    def get_default_quantized_func(self):
        return torch.ops.quantized_decomposed.dequantize_per_channel

    def get_scalar_quantized_func(self):
        return self.get_default_quantized_func()

    def __call__(self, input, scale, zero_point=None, axis=0, quant_min=None, quant_max=None, dtype=torch.float8_e4m3fn, out_dtype=torch.bfloat16):
        return self._quantized_func_(input, scale, zero_point, axis, quant_min, quant_max, dtype=dtype, out_dtype=out_dtype)

_OP_TYPE_XPU_QUANTIZED_WRAPPER_CLASSES = {
                                          OP_TYPE.LINEAR_GEMM : QuantizedXPUMatmul,
                                          OP_TYPE.MATMUL_GEMM : QuantizedXPUMatmul,
                                          OP_TYPE.CAST_TO_FP8 : QuantizedXPUCastToFP8Base,
                                          OP_TYPE.CAST_FROM_FP8 : QuantizedXPUCastFromFP8Base,
                                          OP_TYPE.QUANT: QuantizedXPUQuant,
                                          OP_TYPE.DEQUANT: QuantizedXPUDeQuant,
                                          OP_TYPE.QUANT_PC: QuantizedXPUQuantPC,
                                          OP_TYPE.DEQUANT_PC: QuantizedXPUDeQuantPC,
                                         }


def init_xpu_quantized_func_wrapper_factory():
    QuantizedFuncWrapperFactory.initialize(_OP_TYPE_XPU_QUANTIZED_WRAPPER_CLASSES)
